
#define DEBUG_REGIONS 0
#define DEBUG_INSERTS 0

#include <math.h>
#include "global.h"
#include "processor.h"
#include "list.h"
#include "dag.h"
#include "csp.h"


/*
 *  Internally defined routines.
 */
static void	criticalPathDistances( CSP_type *C, DAG_type *DAG );
static void     setLatencyConstraints( CSP_type *C, DAG_type *DAG );
static void	updateCriticalPath( CSP_type *C );
static BOOL	edgeRedundant( int **cp, int s, int t, int latency );
static void	pruneRedundantConstraints( CSP_type *C,
			ELEM **succList, ELEM **predList );
static void     setDistanceConstraints( CSP_type *C );
static int      setDominanceConstraints( CSP_type *C, int iteration );
static void	setEdgeConstraints( CSP_type *C );

/*
 *  Externally defined routine.
 */
int solve( CSP_type *C, int *length );
int setDominanceConstraints1( CSP_type *C );
int setDominanceConstraints2( CSP_type *C, int s, int t,
					BOOL *inRegion, char ch );

/*
 *  Generate a CSP corresponding to an
 *  instance of the instruction scheduling problem.
 */
void
setConstraints( CSP_type *C, DAG_type *DAG )
{
    int  v, w;
    ELEM *p;
    int  iteration, nConstraints;

    criticalPathDistances( C, DAG );
    setLatencyConstraints( C, DAG );
    setDistanceConstraints( C );

    /*
     *  setDominanceConstraints must come after setDistanceConstraints
     *  since the former checks for cycles before inserting a constraint
     *  and the latter does not. Also, setDom does not assume that
     *  i-->j implies i < j; whereas setDist does.
     */

    set_relative_CPU_limit( C, TIME_LIMIT_SECOND );
    iteration = 0;
    do {
        iteration++;
    
	if( iteration == 1 )
	    nConstraints = setDominanceConstraints1( C );
	else
	    nConstraints = setDominanceConstraints( C, iteration );

#if DEBUG_INSERTS
printf("Iteration %d : %d constraints added : %0.2f seconds\n",
    iteration, nConstraints, elapsed_time(C) );
#endif

	if( nConstraints > 0 ) {
            updateCriticalPath( C );
            pruneRedundantConstraints( C, C->latencySucc, C->latencyPred );
            pruneRedundantConstraints( C, C->distanceSucc, C->distancePred );

            for( v = 1; v <= C->n; v++ ) {
    	        for( p = C->dominanceSucc[v]; p != NULL; p = p->next ) {
    	            w = p->vertex;
    	            if( !edgeRedundant( C->cp, v, w, p->latency ) ) {
    		        insertList( &C->latencySucc[v], w, p->latency );
    		        insertList( &C->latencyPred[w], v, p->latency );
    	            }
    	        }
    	        destroyList( C->dominanceSucc[v] );
    	        destroyList( C->dominancePred[v] );
    	        C->dominanceSucc[v] = NULL;
    	        C->dominancePred[v] = NULL;
            }

            for( v = 1; v <= C->n; v++ ) {
    	        C->nPreds[v] = length( C->latencyPred[v] );
    	        C->nSuccs[v] = length( C->latencySucc[v] );
            }
	}

    } while( ((nConstraints > 20 && iteration < 10) || (iteration == 1)) &&
		!limit_exceeded( C ) );

    //printGraph( C);
    //exit(6);

    setEdgeConstraints( C );
}

/*
 *  Return true if an edge in the DAG is redundant.
 *  Assumption: i < j if there is an edge from i to j.
 *  Sometimes called when the assumption may not hold,
 *  in which case the routine may say the edge is not
 *  redundant when in fact it is.
 */
static BOOL
edgeRedundant( int **cp, int s, int t, int latency )
{
    int  v;
    BOOL path;

    if( latency < cp[s][t] ) {
	return( TRUE );
    }
    else { // latency == cp[s][t]
	path = FALSE;
	for( v = s+1; (v < t) && !path; v++ ) {
	    // there exists a longer path of equal latency
	    if( latency == (cp[s][v] + cp[v][t]) ) {
		path = TRUE;
	    }
	}
	return( path );
    }
}

/*
 *  Set the latency constraints of the CSP.
 *
 *  latencySucc[v] is a list of all nodes w and latencies l
 *	such that there is a latency constraint between v and w
 *	of the form: lower[v] + l <= lower[w].
 *  latencyPred[w] is a list of all nodes v and latencies l
 *	such that there is a latency constraint between w and v
 *	of the form: upper[w] - l >= upper[v].
 */
static void
setLatencyConstraints( CSP_type *C, DAG_type *DAG )
{
    int  u, v, w, latency;
    int  width, nodeCount;
    ELEM *q;
    BOOL changed;

    /*
     *  In the DAG, the vertices are named 0, ..., n-1.
     *  In the CSP, the vertices are named 1, ..., n.
     */
    changed = FALSE;
    for( v = 1; v <= C->n; v++ ) {
	for( q = DAG->successors[v-1]; q != NULL; q = q->next ) {
	    w       = q->vertex + 1;
	    latency = q->latency;	// latency on edge (v,w)
	    /*
	     *  Insert into CSP if not redundant.
	     */
	    if( !edgeRedundant( C->cp, v, w, latency ) ) {
		/*
		 *  Check for special case where latency 0 edge
		 *  can be improved to latency 1.
		 */
		if( (latency == 0) &&
		    (C->varType[v] == C->varType[w]) &&
		    (nFunctionalUnits( C->varType[v] ) == 1) ) {
		    latency = 1;
		    C->cp[v][w] = 1; // Needed for special case below
		    changed = TRUE;
		}
		//printf("%s: %d --%d--> %d\n",DAG->blockID,v,latency,w);
		insertList( &C->latencySucc[v], w, latency );
		insertList( &C->latencyPred[w], v, latency );
	    }
	}
    }

    /*
     *  Improves lots of problems (e.g., in gcc benchmark).
     *  To avoid creating many, many uninteresting regions to
     *  the DAG (and the resulting slow down in setDistance),
     *  insert new edges onto distance lists, rather than
     *  latency lists.
IMPROVEMENT: this should be nodeCount == (width+1), and remove
the (nodeCount <= width) condition. This will reduce the number
of redundant constraints that are added.
Or should it be: remove the (nodeCount <= width) condition and
add an edge with latency  ceil(nodeCount/width) - 1  ???
Also: look at removing the condition type(u) = type(v)
     */
    for( v = 1; v <= C->n; v++ )
    for( w = v+1; w <= C->n; w++ ) {
	if( (C->cp[v][w] == 0) &&
	    (C->varType[v] == C->varType[w]) ) {
	    nodeCount = 2;
	    width = nFunctionalUnits( C->varType[v] );
	    for( u = v+1; (u < w) && (nodeCount <= width); u++ ) {
		if( (C->varType[u] == C->varType[v]) &&
		    (C->cp[v][u] >= 0) && (C->cp[u][w] >= 0) ) {
		    nodeCount++;
		}
	    }
	    if( nodeCount > width ) {
		//insertList( &C->latencySucc[v], w, 1 );
		//insertList( &C->latencyPred[w], v, 1 );
		insertList( &C->distanceSucc[v], w, 1 );
		insertList( &C->distancePred[w], v, 1 );
		changed = TRUE;
	    }
	}
    }

    if( changed ) {
	updateCriticalPath( C );
    }

    for( v = 1; v <= C->n; v++ ) {
	C->nPreds[v] = length( C->latencyPred[v] );
	C->nSuccs[v] = length( C->latencySucc[v] );
    }
}

/*
 *  Return true if the source node s has more than one successor and
 *  the sink node t has more than one predecessor in the region.
 */
static BOOL
validRegion( CSP_type *C, int s, int t )
{
    int   w, count;
    ELEM  *p;

    count = 0;
    for( p = C->latencySucc[s]; p != NULL && count <= 1; p = p->next ) {
	w = p->vertex;
	if( C->cp[w][t] >= 0 ) {
	    count++;
	}
    }
    if( count <= 1 ) {
	return( FALSE );
    }

    count = 0;
    for( p = C->latencyPred[t]; p != NULL && count <= 1; p = p->next ) {
	w = p->vertex;
	if( C->cp[s][w] >= 0 ) {
	    count++;
	}
    }
    if( count <= 1 ) {
	return( FALSE );
    }

    return( TRUE );
}

static int
solveRegion( CSP_type *C, int s, int t, BOOL *inRegion )
{
    CSP_type    *subC;
    int  i, w, n, distance, status, type;
    int  *map;
    ELEM *p;

    map = (int *)calloc( (size_t)C->n+1, sizeof(int) );

    n = 0;
    n++;
    map[s] = n;
    for( w = s+1; w <= t-1; w++ ) {
	if( inRegion[w] ) {
	    n++;
	    map[w] = n;
	}
    }
    n++;
    map[t] = n;

    subC = createCSP( n );
    for( i = s; i <= t; i++ ) if( inRegion[i] ) {
	subC->varType[map[i]] = C->varType[i];
    }
    for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
	subC->nOps[type] = 0;
    }
    for( i = 1; i <= subC->n; i++ ) {
	subC->nOps[subC->varType[i]]++;
    }

    /*
     *  Set latency/distance constraints.
     */
    for( i = s; i < t; i++ ) if( inRegion[i] ) {

	for( p = C->latencySucc[i]; p != NULL; p = p->next ) {
	    w = p->vertex;
	    if( inRegion[w] ) {
		insertList( &subC->latencySucc[map[i]], map[w], p->latency );
		insertList( &subC->latencyPred[map[w]], map[i], p->latency );
	    }
	}

	for( p = C->distanceSucc[i]; p != NULL; p = p->next ) {
	    w = p->vertex;
	    if( inRegion[w] ) {
		insertList( &subC->distanceSucc[map[i]], map[w], p->latency );
		insertList( &subC->distancePred[map[w]], map[i], p->latency );
	    }
	}
    }

    for( i = 1; i <= subC->n; i++ ) {
	subC->nPreds[i] = length( subC->latencyPred[i] );
	subC->nSuccs[i] = length( subC->latencySucc[i] );
    }

//TEMPORARY
    updateCriticalPath( subC );
//END TEMPORARY

    setBounds( subC );
    subC->ub = HUGE; // no useful bound on schedule length known

    /*
     * Returns SOLUTION,  distance is length of optimal solution
     * Returns TIMED_OUT, distance is lower bound on length of optimal solution
     * In both cases it is valid for use in a distance constraint.
     */
    set_relative_CPU_limit( subC, TIME_LIMIT_FIRST );
    subC->propagationLevel = 0;
    subC->restartsAllowed = FALSE;
    status = solve( subC, &distance );

    destroyCSP( subC );
    free( map );

    return( distance-1 );
}

#define MIN_N  3

static int
estimateRegion( CSP_type *C, int s, int t, BOOL *inRegion )
{
    int      i, j, max_i, max_j, max_a, max_b, w;
    int      distance, t_distance, d;
    int	     type, nOpsRegion[N_INSTRUCTION_TYPES];
    int      ms[N_INSTRUCTION_TYPES][MIN_N+1];
    int      ns[N_INSTRUCTION_TYPES][MIN_N+1];
    int      mp[N_INSTRUCTION_TYPES][MIN_N+1];
    int      np[N_INSTRUCTION_TYPES][MIN_N+1];

    for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
	for( j = 0; j <= MIN_N; j++ ) {
	    ms[type][j] = HUGE;
	    ns[type][j] = 0;
	    mp[type][j] = HUGE;
	    np[type][j] = 0;
	}
	nOpsRegion[type] = 0;
    }

    /*
     *  Count the nodes w that are necessarily
     *  between s and t (i.e., on a path from s to t).
     *  Determine the minimum critical path distances.
     */
    for( w = s; w <= t; w++ ) if( inRegion[w] ) {
	type = C->varType[w];
	nOpsRegion[type]++;
	if( C->cp[s][w] <= ms[type][MIN_N] ) {
	    i = 0;
	    while( C->cp[s][w] > ms[type][i] ) {
		i++;
	    }
	    if( C->cp[s][w] == ms[type][i] ) {
		for( j = MIN_N; j > i; j-- ) {
		    ns[type][j]++;
		}
	    }
	    else
	    if( C->cp[s][w] < ms[type][i] ) {
		for( j = MIN_N; j > i; j-- ) {
		    ms[type][j] = ms[type][j-1];
		    ns[type][j] = ns[type][j-1]+1;
		}
		ms[type][i] = C->cp[s][w];
	    }
	}

	if( C->cp[w][t] <= mp[type][MIN_N] ) {
	    i = 0;
	    while( C->cp[w][t] > mp[type][i] ) {
		i++;
	    }
	    if( C->cp[w][t] == mp[type][i] ) {
		for( j = MIN_N; j > i; j-- ) {
		    np[type][j]++;
		}
	    }
	    else
	    if( C->cp[w][t] < mp[type][i] ) {
		for( j = MIN_N; j > i; j-- ) {
		    mp[type][j] = mp[type][j-1];
		    np[type][j] = np[type][j-1]+1;
		}
		mp[type][i] = C->cp[w][t];
	    }
	}
    }

    /*
     *  Lower bound on the distance between s and t.
     */
    distance = 0;
    for( type = 0; type < N_INSTRUCTION_TYPES; type++ )  {
	if( nOpsRegion[type] != 0 ) {

	    max_a = 0;
	    max_i = 0;
	    for( i = 0; i <= MIN_N; i++ ) if( ms[type][i] != HUGE ) {
		d = ms[type][i] + (int)ceil(((double)nOpsRegion[type]-ns[type][i]) /
					    ((double)nFunctionalUnits( type )));
		if( max_a < d ) {
		    max_a = d;
		    max_i = i;
		}
	    }

	    max_b = 0;
	    max_j = 0;
	    for( i = 0; i <= MIN_N; i++ ) if( mp[type][i] != HUGE && nOpsRegion[type]-ns[type][max_i]-np[type][i]>0 ) {
		d = mp[type][i] + (int)ceil(((double)nOpsRegion[type]-np[type][i]) /
					    ((double)nFunctionalUnits( type )));

		if( max_b < d ) {
		    max_b = d;
		    max_j = i;
		}
	    }

	    t_distance =
		ms[type][max_i] +
		(int)ceil(((double)nOpsRegion[type]-ns[type][max_i]-np[type][max_j]) /
			 ((double)nFunctionalUnits( type ))) +
		mp[type][max_j] +
		- 1;
	    if( t_distance > distance ) distance = t_distance;
	}
    }

    return( distance );
}

/*
 *  Set the distance constraints of the CSP.
 *
 *  distanceSucc[s] is a list of all nodes t and distances d
 *	such that there is a distance constraint between s and t
 *	of the form: lower[s] + d <= lower[t].
 *  distancePred[t] is a list of all nodes s and distances d
 *	such that there is a distance constraint between t and s
 *	of the form: upper[t] - d >= upper[s].
 *
 *  There is potentially a distance constraint between two nodes
 *  s and t if there is there is more than one path from s to t.
 *  If the region is small enough, it is solved exactly; otherwise
 *  the distance is estimated.
 */
static void
setDistanceConstraints( CSP_type *C )
{
    int   s, t, w, distance, nodeCount, latest;
    BOOL  *inRegion;
    ELEM  *p;

    inRegion = (BOOL *)calloc( (size_t)C->n+1, sizeof(BOOL) );

    for( s = C->n; s >= 1; s-- ) if( C->nSuccs[s] > 1 ) {

	/*
	 *  Look for the first partition node between s and the sink
	 *  node (C->n). We are guaranteed to find one as the sink
	 *  node is a partition node.
	 */
	latest = 0;
	for( w = s; w <= C->n; w++ ) {
	    if( (w == s) || (C->cp[s][w] >= 0) ) {
		if( w == latest ) {
		    break;
		}
		for( p = C->latencySucc[w]; p != NULL; p = p->next ) {
		    if( p->vertex > latest ) {
			latest = p->vertex;
		    }
		}
	    }
	}

	/*
	 *  Consider the region with root s and sink t if
	 *  (a) s has more than one successor,
	 *  (b) t has more than one predecessor, and
	 *  (c) there is a path from s to t.
	 */
	for( t = s+1; t <= latest; t++ ) if( (C->nPreds[t] > 1) &&
					     (C->cp[s][t] >= 0) &&
					     validRegion( C, s, t ) ) {
	    /*
	     *  Mark and count the nodes w that are necessarily
	     *  between s and t (i.e., on a path from s to t).
	     */
	    for( w = 1; w <= C->n; w++ ) {
		inRegion[w] = FALSE;
	    }
	    inRegion[s] = TRUE;
	    inRegion[t] = TRUE;
	    nodeCount = 0;
	    for( w = s+1; w < t; w++ ) {
		if( (C->cp[s][w] >= 0) && (C->cp[w][t] >= 0) ) {
		    inRegion[w] = TRUE;
		    nodeCount++;
		}
	    }

	    if( nodeCount <= 2 ) {
		continue;
	    }

	    /*
	     *  If the region is small enough, find the optimal value of
	     *  the distance between s and t; otherwise estimate it.
	     */
	    if( (nodeCount <= MAX_REGION_SIZE) && (C->n > MAX_REGION_SIZE) ) {
		distance = solveRegion( C, s, t, inRegion );
	    }
	    else {
		distance = estimateRegion( C, s, t, inRegion );
	    }

	    /*
	     *  Insert distance constraint into the CSP if it is an
	     *  improvement over critical path distance.
	     */
	    if( distance > C->cp[s][t] ) {
		insertList( &C->distanceSucc[s], t, distance );
		insertList( &C->distancePred[t], s, distance );
	    }

	} /* for t if ... */
    } /* for s */

    free( inRegion);
}


static int
setDominanceConstraints( CSP_type *C, int iteration )
{
    int   s, t, w, latest;
    int   nConstraints, totalConstraints;
    BOOL  *inRegion;
    ELEM  *p;

    inRegion = (BOOL *)calloc( (size_t)C->n+1, sizeof(BOOL) );

    totalConstraints = 0;

    for( s = C->n; s >= 1; s-- ) if( C->nSuccs[s] > 1 ) {

	latest = 0;
	for( w = s+1; w <= C->n; w++ ) {
	    if( C->varType[w] == BRANCH_INSTR ) {
		latest = w;
		break;
	    }
	}
	if( latest == 0 ) latest = C->n;

	/*
	 *  Consider the region with root s and sink t if
	 *  (a) s has more than one successor,
	 *  (b) t has more than one predecessor, and
	 *  (c) there is a path from s to t.
	 */

#if DEBUG_REGIONS
printf("Considering regions (%d, *) where latest = %d\n",s-1,latest-1);
for( t = s+1; t <= latest; t++ ) {
    printf("t = %d, nPreds[%d] = %d, C->cp[%d][%d] = %d, validRegion( C, %d, %d ) = %d\n",
	t-1,t-1,C->nPreds[t],s-1,t-1,C->cp[s][t],s-1,t-1,validRegion( C, s, t ));
    for( p = C->latencyPred[t]; p != NULL; p = p->next ) {
        printf("(%d,%d):%d\n",p->vertex-1,t-1,p->latency);
    }
}
#endif

	for( t = s+1; t <= latest; t++ ) if( (C->nPreds[t] > 1) &&
					     (C->cp[s][t] >= 0) &&
					     validRegion( C, s, t ) ) {

#if DEBUG_REGIONS
printf("Considering region (%d, %d)\n",s-1,t-1);
#endif

	    if( limit_exceeded( C ) ) {
		return(totalConstraints);
	    }

	    //a. %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	    for( w = 1; w <= C->n; w++ ) {
		inRegion[w] = FALSE;
	    }
	    inRegion[s] = TRUE;
	    inRegion[t] = TRUE;
	    for( w = 1; w <= C->n; w++ ) {
		if( (C->cp[s][w] >= 0) && (C->cp[w][t] >= 0) ) {
		    if( C->nPreds[w] <= 3 && C->nSuccs[w] <= 3 ) {
			inRegion[w] = TRUE;
		    }
		}
	    }
	    nConstraints = setDominanceConstraints2( C, s, t, inRegion, 'a' );
	    totalConstraints += nConstraints;

	    //b. %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	    if( nConstraints ) continue;
	    for( w = 1; w <= C->n; w++ ) {
		inRegion[w] = FALSE;
	    }
	    inRegion[s] = TRUE;
	    inRegion[t] = TRUE;
	    for( w = 1; w <= C->n; w++ ) {
		if( (C->cp[s][w] >= 0) && (C->cp[w][t] >= 0) ) {
		    inRegion[w] = TRUE;
		}
	    }
	    nConstraints = setDominanceConstraints2( C, s, t, inRegion, 'b' );
	    totalConstraints += nConstraints;

	    //c. %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	    if( nConstraints ) continue;
	    for( w = 1; w <= C->n; w++ ) {
		inRegion[w] = FALSE;
	    }
	    inRegion[s] = TRUE;
	    inRegion[t] = TRUE;
	    for( p = C->latencySucc[s]; p != NULL; p = p->next ) {
		w = p->vertex;
		if( (C->cp[s][w] >= 0) && (C->cp[w][t] >= 0) ) {
		    inRegion[w] = TRUE;
		}
	    }
	    for( w = 1; w <= C->n; w++ ) {
		if( (C->cp[s][w] >= 0) && (C->cp[w][t] >= 0) ) {
		    if( C->nPreds[w] <= 3 && C->nSuccs[w] <= 3 ) {
		        inRegion[w] = TRUE;
		    }
		}
	    }
	    nConstraints = setDominanceConstraints2( C, s, t, inRegion, 'c' );
	    totalConstraints += nConstraints;

	    //d. %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
	    if( nConstraints ) continue;
	    for( w = 1; w <= C->n; w++ ) {
		inRegion[w] = FALSE;
	    }
	    inRegion[s] = TRUE;
	    inRegion[t] = TRUE;
	    for( p = C->latencyPred[t]; p != NULL; p = p->next ) {
		w = p->vertex;
		if( (C->cp[s][w] >= 0) && (C->cp[w][t] >= 0) ) {
		    inRegion[w] = TRUE;
		}
	    }
	    for( w = 1; w <= C->n; w++ ) {
		if( (C->cp[s][w] >= 0) && (C->cp[w][t] >= 0) ) {
		    if( C->nPreds[w] <= 3 && C->nSuccs[w] <= 3 ) {
		        inRegion[w] = TRUE;
		    }
		}
	    }
	    nConstraints = setDominanceConstraints2( C, s, t, inRegion, 'd' );
	    totalConstraints += nConstraints;

	} /* for t if ... */
    } /* for s */

    free( inRegion );

    return( totalConstraints );
}

/*
 *  Set the edge constraints of the CSP.
 *
 *  edgePred[v] is a list of all nodes w such that w is a
 *	successor of v and w has more than one predecessor.
 *	There is an edge constraint between w and v and the
 *	other predecessors of w.
 *  edgeSucc[w] is a list of all nodes v such that v is a
 *	predecessor of w and v has more than one successor.
 *	There is an edge constraint between v and w and the
 *	other successors of v.
 */
static void
setEdgeConstraints( CSP_type *C )
{
    int  v;
    ELEM *p;

    for( v = 1; v <= C->n; v++ ) {
	if( C->nPreds[v] > 1 ) {
	    for( p = C->latencyPred[v]; p != NULL; p = p->next ) {
		insertList( &C->edgePred[p->vertex], v, 0/* not used */ );
	    }
	}
    }

    for( v = 1; v <= C->n; v++ ) {
	if( C->nSuccs[v] > 1 ) {
	    for( p = C->latencySucc[v]; p != NULL; p = p->next ) {
		insertList( &C->edgeSucc[p->vertex], v, 0/* not used */ );
	    }
	}
    }
}

static void
pruneRedundantConstraints( CSP_type *C, ELEM **succList, ELEM **predList )
{
    int i, v, w;
    ELEM **saveList, *p;

    saveList = (ELEM **) calloc( (size_t)C->n+1, sizeof(ELEM *) );
    for( i = 1; i <= C->n; i++ ) {
	saveList[i] = succList[i];
    }

    for( i = 1; i <= C->n; i++ ) {
	destroyList( predList[i] );
        succList[i] = NULL;
        predList[i] = NULL;
    }

    for( v = 1; v <= C->n; v++ ) {
	for( p = saveList[v]; p != NULL; p = p->next ) {
	    w = p->vertex;
	    if( !edgeRedundant( C->cp, v, w, p->latency ) ) {
		insertList( &succList[v], w, p->latency );
		insertList( &predList[w], v, p->latency );
	    }
	}
    }

    for( i = 1; i <= C->n; i++ ) {
	destroyList( saveList[i] );
    }
    free( saveList );
}

/*
 *  Determine critical path distances for each vertex.
 *  Assumption: all successors of a vertex have a higher vertex number.
 */
static void
criticalPathDistances( CSP_type *C, DAG_type *DAG )
{
    int  s, u, v;
    ELEM *q;

    for( s = 1; s <= C->n; s++ ) {
	for( v = 1; v <= C->n; v++ ) {
	    C->cp[s][v] = -HUGE;
	    C->edgeMarked[s][v] = FALSE;
	}
	C->cp[s][s] = 0;
	C->edgeMarked[s][s] = TRUE;
    }

    for( s = 1; s <= C->n; s++ )
    for( u = s; u <= C->n; u++ ) {
        for( q = DAG->successors[u-1]; q != NULL; q = q->next ) {
	    v = q->vertex + 1;
	    if( C->cp[s][v] < (C->cp[s][u] + q->latency) ) {
		C->cp[s][v] = (C->cp[s][u] + q->latency);
	    }
	}
    }
}

static void
topologicalSort(
	CSP_type *C,
	int v,
	int *order,
	int *index,
	BOOL *marked )
{
    int   w;
    ELEM  *p;

    /*
     *  For each successor w of v that is not already marked...
     */
    for( p = C->latencySucc[v]; p != NULL; p = p->next ) {
	w = p->vertex;
	if( !marked[w] ) {
	    topologicalSort( C, w, order, index, marked );
	}
    }
    for( p = C->distanceSucc[v]; p != NULL; p = p->next ) {
	w = p->vertex;
	if( !marked[w] ) {
	    topologicalSort( C, w, order, index, marked );
	}
    }
    for( p = C->dominanceSucc[v]; p != NULL; p = p->next ) {
	w = p->vertex;
	if( !marked[w] ) {
	    topologicalSort( C, w, order, index, marked );
	}
    }

    /*
     *  Insert the node into the order.
     */
    marked[v] = TRUE;
    order[*index] = v;
    (*index)--;
}


/*
 *  Determine critical path distances for each vertex.
 *  Cannot make the assumption that all successors of a vertex have a
 *  higher vertex number as this isn't necessarily true after dominance
 *  constraints have been added. Fixing this required processing the
 *  vertices in topological sorted order.
 */
static void
updateCriticalPath( CSP_type *C )
{
    int  i, j, s, u, v;
    int  *order;
    int  index;
    BOOL *marked;
    ELEM *q;

    order = (int *)calloc( (size_t)C->n+1, sizeof(int) );
    marked = (BOOL *)calloc( (size_t)C->n+1, sizeof(BOOL) );

    index = C->n;
    topologicalSort( C, 1, order, &index, marked );

    for( s = 1; s <= C->n; s++ ) {
	for( v = 1; v <= C->n; v++ ) {
	    C->cp[s][v] = -HUGE;
	    C->edgeMarked[s][v] = FALSE;
	}
	C->cp[s][s] = 0;
	C->edgeMarked[s][s] = TRUE;
    }

    for( i = 1; i <= C->n; i++ )
    for( j = i; j <= C->n; j++ ) {
	s = order[i];
	u = order[j];
        for( q = C->latencySucc[u]; q != NULL; q = q->next ) {
	    v = q->vertex;
	    if( C->cp[s][v] < (C->cp[s][u] + q->latency) ) {
		C->cp[s][v] = (C->cp[s][u] + q->latency);
	    }
	}
        for( q = C->distanceSucc[u]; q != NULL; q = q->next ) {
	    v = q->vertex;
	    if( C->cp[s][v] < (C->cp[s][u] + q->latency) ) {
		C->cp[s][v] = (C->cp[s][u] + q->latency);
	    }
	}
        for( q = C->dominanceSucc[u]; q != NULL; q = q->next ) {
	    v = q->vertex;
	    if( C->cp[s][v] < (C->cp[s][u] + q->latency) ) {
		C->cp[s][v] = (C->cp[s][u] + q->latency);
	    }
	}
    }

    free( order );
    free( marked );
}

