
#include "global.h"
#include "processor.h"
#include "list.h"
#include "dag.h"
#include "csp.h"


/* =================================================================
 *  Routines for creating and destroying a CSP with n variables.
 */

/*
 *  Allocate space for n variables.
 */
CSP_type *
createCSP( int n )
{
    int       i;
    CSP_type  *C;

    C = (CSP_type *)calloc( (size_t)1, sizeof(CSP_type) );

    C->n = n;
    C->visits = 0;
    C->backtracks = 0;

    C->varType = (int *)calloc( (size_t)n+1, sizeof(int) );
    C->nOps    = (int *)calloc( (size_t)N_INSTRUCTION_TYPES+1, sizeof(int) );
    C->lower   = (int *)calloc( (size_t)n+1, sizeof(int) );
    C->upper   = (int *)calloc( (size_t)n+1, sizeof(int) );

    C->cp = (int **)calloc( (size_t)C->n+1, sizeof(int *) );
    for( i = 1; i <= C->n; i++ ) {
	C->cp[i] = (int *)calloc( (size_t)C->n+1, sizeof(int) );
    }

    C->edgeMarked = (BOOL **)calloc( (size_t)C->n+1, sizeof(BOOL *) );
    for( i = 1; i <= C->n; i++ ) {
	C->edgeMarked[i] = (BOOL *)calloc( (size_t)C->n+1, sizeof(BOOL) );
    }

    C->latencySucc = (ELEM **) calloc( (size_t)n+1, sizeof(ELEM *) );
    C->latencyPred = (ELEM **) calloc( (size_t)n+1, sizeof(ELEM *) );
    for( i = 1; i <= n; i++ ) {
	C->latencySucc[i] = (ELEM *)NULL;
	C->latencyPred[i] = (ELEM *)NULL;
    }

    C->distanceSucc = (ELEM **) calloc( (size_t)n+1, sizeof(ELEM *) );
    C->distancePred = (ELEM **) calloc( (size_t)n+1, sizeof(ELEM *) );
    for( i = 1; i <= n; i++ ) {
	C->distanceSucc[i] = (ELEM *)NULL;
	C->distancePred[i] = (ELEM *)NULL;
    }

    C->dominanceSucc = (ELEM **) calloc( (size_t)n+1, sizeof(ELEM *) );
    C->dominancePred = (ELEM **) calloc( (size_t)n+1, sizeof(ELEM *) );
    for( i = 1; i <= n; i++ ) {
	C->dominanceSucc[i] = (ELEM *)NULL;
	C->dominancePred[i] = (ELEM *)NULL;
    }

    C->edgeSucc = (ELEM **) calloc( (size_t)n+1, sizeof(ELEM *) );
    C->edgePred = (ELEM **) calloc( (size_t)n+1, sizeof(ELEM *) );
    for( i = 1; i <= n; i++ ) {
	C->edgeSucc[i] = (ELEM *)NULL;
	C->edgePred[i] = (ELEM *)NULL;
    }

    C->nPreds  = (int *)calloc( (size_t)n+1, sizeof(int) );
    C->nSuccs  = (int *)calloc( (size_t)n+1, sizeof(int) );
    C->changev = (int *)calloc( (size_t)n+1, sizeof(int) );

    C->saveLower = (int **)calloc( (size_t)C->n+1, sizeof(int *) );
    C->saveUpper = (int **)calloc( (size_t)C->n+1, sizeof(int *) );
    for( i = 0; i <= C->n; i++ ) {
	C->saveLower[i] = (int *)calloc( (size_t)C->n+1, sizeof(int) );
	C->saveUpper[i] = (int *)calloc( (size_t)C->n+1, sizeof(int) );
    }

    C->s_lower = createStack( n+1 );
    C->s_upper = createStack( n+1 );

    C->values = (node_type *)calloc( (size_t)n+1, sizeof(node_type) );

    C->filled = (BOOL *)calloc( (size_t)MAX_LATENCY*(n+1), sizeof(BOOL) );

    C->iv        = (interval  *)calloc((size_t)n, sizeof(interval  ));
    C->minsorted = (interval **)calloc((size_t)n, sizeof(interval *));
    C->maxsorted = (interval **)calloc((size_t)n, sizeof(interval *));
    C->bounds    = (int *)calloc((size_t)2*n+2, sizeof(int));

    C->t = (int *)calloc((size_t)2*n+2, sizeof(int));
    C->d = (int *)calloc((size_t)2*n+2, sizeof(int));
    C->h = (int *)calloc((size_t)2*n+2, sizeof(int));

    set_start_time( C );

    return( C );
}

void
destroyCSP( CSP_type *C )
{
    int  i;

    free( C->varType );
    free( C->nOps );
    free( C->lower );
    free( C->upper );

    for( i = 1; i <= C->n; i++ ) {
	free( C->cp[i] );
    }
    free( C->cp );

    for( i = 1; i <= C->n; i++ ) {
	free( C->edgeMarked[i] );
    }
    free( C->edgeMarked );

    for( i = 1; i <= C->n; i++ ) {
	destroyList( C->latencySucc[i] );
	destroyList( C->latencyPred[i] );
    }
    free( C->latencySucc );
    free( C->latencyPred );

    for( i = 1; i <= C->n; i++ ) {
	destroyList( C->distanceSucc[i] );
	destroyList( C->distancePred[i] );
    }
    free( C->distanceSucc );
    free( C->distancePred );

    for( i = 1; i <= C->n; i++ ) {
	destroyList( C->dominanceSucc[i] );
	destroyList( C->dominancePred[i] );
    }
    free( C->dominanceSucc );
    free( C->dominancePred );

    for( i = 1; i <= C->n; i++ ) {
	destroyList( C->edgeSucc[i] );
	destroyList( C->edgePred[i] );
    }
    free( C->edgeSucc );
    free( C->edgePred );

    free( C->nPreds );
    free( C->nSuccs );
    free( C->changev );

    for( i = 0; i <= C->n; i++ ) {
	free( C->saveLower[i] );
	free( C->saveUpper[i] );
    }
    free( C->saveLower );
    free( C->saveUpper );

    destroyStack( C->s_lower );
    destroyStack( C->s_upper );

    free( C->values );

    free( C->filled );

    free( C->iv );
    free( C->minsorted );
    free( C->maxsorted );
    free( C->bounds );

    free( C->h );
    free( C->d );
    free( C->t );

    free( C );
}


/* =================================================================
 *  Routines for manipulating the bounds of the CSP variables.
 */

static void
topologicalSort(
	CSP_type *C,
	int v,
	int *order,
	int *index,
	BOOL *marked )
{
    int   w;
    ELEM  *p;

    /*
     *  For each successor w of v that is not already marked...
     */
    for( p = C->latencySucc[v]; p != NULL; p = p->next ) {
	w = p->vertex;
	if( !marked[w] ) {
	    topologicalSort( C, w, order, index, marked );
	}
    }
    for( p = C->distanceSucc[v]; p != NULL; p = p->next ) {
	w = p->vertex;
	if( !marked[w] ) {
	    topologicalSort( C, w, order, index, marked );
	}
    }

    /*
     *  Insert the node into the order.
     */
    marked[v] = TRUE;
    order[*index] = v;
    (*index)--;
}


/*
 *  Set and improve the lower bounds of the variables by
 *  propagating the latency, distance, and edge constraints.
 */
static void
setLowerBounds( CSP_type *C, int order[] )
{
    int  i, v, w, type;
    ELEM *p;

    for( v = 1; v <= C->n; v++ ) {
	C->lower[v] = 1;
    }

    for( i = 1; i <= C->n; i++ ) {
	v = order[i];

	if( C->nPreds[v] > 1 ) {
	    for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
		doPredsOfNode( C, v, /* pushFlag */FALSE,
					type, nFunctionalUnits( type ) );
	    }
	}

	for( p = C->latencySucc[v]; p != NULL; p = p->next ) {
	    w = p->vertex;
	    if( C->lower[w] < (C->lower[v] + p->latency) ) {
		C->lower[w] =  C->lower[v] + p->latency;
	    }
	}

	for( p = C->distanceSucc[v]; p != NULL; p = p->next ) {
	    w = p->vertex;
	    if( C->lower[w] < (C->lower[v] + p->latency) ) {
		C->lower[w] =  C->lower[v] + p->latency;
	    }
	}
    }
}

/*
 *  Set and improve the upper bounds of the variables by
 *  propagating the latency, distance, and edge constraints.
 */
static void
setUpperBounds( CSP_type *C, int order[], int m )
{
    int  i, v, w, type;
    ELEM *p;

    C->m = m;
    for( v = C->n; v >= 1; v-- ) {
	C->upper[v] = m;
    }

    for( i = C->n; i >= 1; i-- ) {
	v = order[i];

	if( C->nSuccs[v] > 1 ) {
	    for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
		doSuccsOfNode( C, v, /* pushFlag */FALSE,
					type, nFunctionalUnits( type ) );
	    }
	}

	for( p = C->latencyPred[v]; p != NULL; p = p->next ) {
	    w = p->vertex;
	    if( C->upper[w] > (C->upper[v] - p->latency) ) {
		C->upper[w] =  C->upper[v] - p->latency;
	    }
	}
	for( p = C->distancePred[v]; p != NULL; p = p->next ) {
	    w = p->vertex;
	    if( C->upper[w] > (C->upper[v] - p->latency) ) {
		C->upper[w] =  C->upper[v] - p->latency;
	    }
	}
    }
}

/*
 *  Increment each of the upper bounds of the variables,
 *  ensuring that the first and the last variable have
 *  their domains equal to [1,1] and [m,m], respectively.
 */
void
incrementUpperBounds( CSP_type *C )
{
    int  v;

    C->m++;
    for( v = 1; v <= C->n; v++ ) {
	C->upper[v]++;
    }

    C->lower[1]    = C->upper[1]    = 1;
    C->lower[C->n] = C->upper[C->n] = C->m;
}

/*
 *  Set the initial lower and upper bounds of the variables.
 *  ensuring that the first and the last variable have
 *  their domains equal to [1,1] and [m,m], respectively.
 */
void
setBounds( CSP_type *C )
{
    int  m;
    int  *order;
    int  index;
    BOOL *marked;

    order = (int *)calloc( (size_t)C->n+1, sizeof(int) );
    marked = (BOOL *)calloc( (size_t)C->n+1, sizeof(BOOL) );

    index = C->n;
    topologicalSort( C, 1, order, &index, marked );

    setLowerBounds( C, order );

    m = C->lower[C->n];
    setUpperBounds( C, order, m );

    C->lower[1] = C->upper[1] = 1;
    C->lower[C->n] = C->upper[C->n] = m;

    free( order );
    free( marked );
}

/*
 *  Save the lower and upper bounds of the variables at
 *  currentLevel.
 */
void
saveBounds( CSP_type *C, int currentLevel )
{
    int  v;

    for( v = 1; v <= C->n; v++ ) {
	C->saveLower[currentLevel][v] = C->lower[v];
	C->saveUpper[currentLevel][v] = C->upper[v];
    }
}

/*
 *  Restore the lower and upper bounds of the variables at
 *  currentLevel.
 */
void
restoreBounds( CSP_type *C, int currentLevel )
{
    int  v;

    for( v = 1; v <= C->n; v++ ) {
	C->lower[v] = C->saveLower[currentLevel][v];
	C->upper[v] = C->saveUpper[currentLevel][v];
    }
}

/* =================================================================
 *  Routines for printing/debugging/viewing.
 */

void
printCSP( CSP_type *C )
{
    int  v;
    ELEM *p;

    printf( "\tn: %d\n", C->n );
    printf( "\tm: %d\n\n",  C->m );
    printf( "\tub: %d\n\n",  C->ub );
    for( v = 1; v <= C->n; v++ ) {
	printf( "\t%3d: [%3d,%3d] = %3d; pred=%3d; succ=%3d; type=%d\n",
	    v,
	    C->lower[v],
	    C->upper[v],
	    C->upper[v] - C->lower[v] + 1,
	    C->nPreds[v],
	    C->nSuccs[v],
	    C->varType[v] );
    }
    printf( "\n" );

    printf( "Successors:\n" );
    for( v = 1; v <= C->n; v++ ) {
	printf( "%d", v );
	for( p = C->latencySucc[v]; p != NULL; p = p->next ) {
	    printf( " %d:%d", p->vertex, p->latency );
	}
	printf( "\n" );
    }
    printf( "\n" );

    printf( "Predecessors:\n" );
    for( v = 1; v <= C->n; v++ ) {
	printf( "%d", v );
	for( p = C->latencyPred[v]; p != NULL; p = p->next ) {
	    printf( " %d:%d", p->vertex, p->latency );
	}
	printf( "\n" );
    }
    printf( "\n" );
}

void
printSolution( CSP_type *C )
{
    int  t, v;

    for( t = 1; t <= C->m; t++ ) {
	printf( "Instruction(s) at time %d:", t );
	for( v = 1; v <= C->n; v++ ) {
	    if( C->lower[v] == t && C->upper[v] == t ) {
		printf( " %d", v-1 );
	    }
	}
	printf( "\n" );
    }
}

/*
 *  Print the CSP graph in a form suitable for viewing by yEd.
 */
void
printGraph( CSP_type *C )
{
    int	 v;
    ELEM *q;

    printf("graph\n");
    printf("[\n");
    printf("\tdirected\t1\n");

    for( v = 1; v <= C->n; v++ ) {
	printf("\tnode\n");
	printf("\t[\n");
	printf("\t\tid\t%d\n", v-1);
	printf("\t\tlabel\t\"%d\"\n", v-1);
	printf("\t\tgraphics\n");
	printf("\t\t[\n");
	printf("\t\t\ttype\t\"rectangle\"\n");
	switch( C->varType[v] ) {
	    case 0: printf("\t\t\tfill\t\"#FF0000\"\n");
		    break;
	    case 1: printf("\t\t\tfill\t\"#8EFF00\"\n");
		    break;
	    case 2: printf("\t\t\tfill\t\"#002DFF\"\n");
		    break;
	    case 3: printf("\t\t\tfill\t\"#FFFF00\"\n");
		    break;
	    case 4: printf("\t\t\tfill\t\"#FFFFFF\"\n");
		    break;
	}
	printf("\t\t]\n");
	printf("\t]\n");
    }

    for( v = 1; v <= C->n; v++ ) {
	for( q = C->latencySucc[v]; q != NULL; q = q->next ) {
	    printf("\tedge\n");
	    printf("\t[\n");
	    printf("\t\tsource\t%d\n", v-1);
	    printf("\t\ttarget\t%d\n", q->vertex-1);
	    printf("\t\tlabel\t\"%d\"\n", q->latency);
	    printf("\t\tgraphics\n");
	    printf("\t\t[\n");
	    printf("\t\t\tfill\t\"#000000\"\n");
	    printf("\t\t\ttargetArrow\t\"standard\"\n");
	    printf("\t\t]\n");
	    printf("\t]\n");
	}
    }
    printf("]\n");
    fflush(stdout);
}

void
verifySolution( CSP_type *C )
{
    int  t, v, type, status, *count;
    ELEM *p;

    count = (int *)calloc( (size_t)C->m+1, sizeof(int) );

    status = 1;
    for( v = 1; v <= C->n; v++ ) {
	status = status && (C->lower[v] == C->upper[v]);
    }
    if( !status ) {
	printf( "Verifying CSP Scheduler solution:\n" );
	printf( "\tEach instruction does not have a unique value.\n" );
	exit( 6 );
    }

    for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
	for( t = 1; t <= C->m; t++ ) {
	    count[t] = 0;
	}
	for( v = 1; v <= C->n; v++ ) {
	    if( C->varType[v] == type )
		count[C->lower[v]]++;
	}
	status = 1;
	for( t = 1; t <= C->m; t++ ) {
	    status = status && (count[t] <= nFunctionalUnits( type ));
	}
	if( !status ) {
	    printf( "Verifying CSP Scheduler solution:\n" );
	    printf( "\tIssue-width constraint on type %d not satisfied.\n", type );
	    exit( 6 );
	}
    }

    status = 1;
    for( v = 1; v <= C->n; v++ ) {
	for( p = C->latencySucc[v]; p != NULL; p = p->next ) {
	    status = status &&
			((C->lower[v] + p->latency) <= C->lower[p->vertex]);
	}
    }
    if( !status ) {
	printf( "Verifying CSP Scheduler solution:\n" );
	printf( "\tSuccessor constraints not satisfied.\n" );
	exit( 6 );
    }

    status = 1;
    for( v = 1; v <= C->n; v++ ) {
	for( p = C->latencyPred[v]; p != NULL; p = p->next ) {
	    status = status &&
			((C->lower[p->vertex] + p->latency) <= C->lower[v]);
	}
    }
    if( !status ) {
	printf( "Verifying CSP Scheduler solution:\n" );
	printf( "\tPredecessor constraints not satisfied.\n" );
	exit( 6 );
    }

    free( count );
}

