
#include "stack.h"
#include "propagate.h"

/* =================================================================
 *  Abstract data type to represent a CSP.
 */
typedef struct {
    /*
     *  Variables.
     */
    int  n;		/* number of variables; their names are 1,...,n */
    /*
     *  Properties.
     */
    int  *varType;	/* operation type of each variable/instruction	*/
    int  *nOps;		/* number of operations of each type		*/
    int  *nPreds;	/* number of immediate predecessors of v	*/
    int  *nSuccs;	/* number of immediate successors of v		*/
    int  **cp;		/* critical path distances between variables	*/
    BOOL **edgeMarked;
    /*
     *  Domains.
     */
    int  m;		/* domains are in the range [1, ..., m]		*/
    int  *lower;	/* current lower bound on variable v		*/
    int  *upper;	/* current upper bound on variable v		*/
    int  ub;		/* upper bound on schedule length		*/
    /*
     *  Constraints.
     *  For a description of a constraint, see the comments
     *  for the corresponding routine that sets that constraint.
     */
    ELEM **latencySucc;
    ELEM **latencyPred;
    ELEM **distanceSucc;
    ELEM **distancePred;
    ELEM **dominanceSucc;
    ELEM **dominancePred;
    ELEM **edgeSucc;
    ELEM **edgePred;
    /*
     *  Solving/search.
     */
    /*
     *  Statistics.
     */
    int	 visits;	/* Nodes visited so far during search		*/
    int	 backtracks;	/* Backtracks performed so far during search	*/
    /*
     *  Saving/restoring.
     *  Space for storing (and restoring) the lower and upper bounds
     *  of the variables at the 0, ..., n levels of the search.
     */
    int  **saveLower;
    int  **saveUpper;
    /*
     *  Variable ordering heuristic.
     */
    int  *changev;	/* amount of propagation this variable caused when */
			/*     probing was used to test lower bound        */
    /*
     *  Propagating.
     */
    int		propagationLevel;/* 0 means propagate, > 0 means probing */
    int		changes;	/* domains changed during last propagation */
    BOOL	iterNeeded[N_INSTRUCTION_TYPES];
    STACK_t	*s_lower;
    STACK_t	*s_upper;
    node_type	*values;	/* used by doPred, doSucc */
    BOOL	*filled;	/* used by BinaryGCC	  */
    int		*t;		/* used by gcc ...	  */
    int		*d;
    int		*h;
    interval	*iv;
    int		niv;
    interval	**minsorted;
    interval	**maxsorted;
    int		*bounds;
    int		nb;
    /*
     *  Timing and restarts.
     */
    double	startTime;
    double	maxCPUseconds;
    BOOL	restartsAllowed;
} CSP_type;

/*
 *  Possible values returned by solve and probe routines.
 */
#define	  NO_SOLUTION	  0	/* no solution exists		 	 */
#define	  PROVED_OPTIMAL  1	/* heuristic solution was proven optimal */
#define	  SOLUTION	  2	/* a solution was found			 */
#define	  TIMED_OUT	  3	/* unknown as resource limit exceeded 	 */

/*
 *  Timeouts for two stages of solving (without probing, with probing).
 */
#define   TIME_LIMIT_FIRST	5.0  /* 1st limit on CPU time (seconds)   */
#define   TIME_LIMIT_SECOND    15.0  /* 2nd limit on CPU time (seconds)   */
#define   TIME_LIMIT_TOTAL    600.0  /* total limit on CPU time (seconds) */

/*
 *  If the number of nodes necessarily between two nodes s and v
 *  is less than or equal to this value, then attempt to improve
 *  the distance estimate between these nodes.
 */
#define MAX_REGION_SIZE		25

/*
 *  For problems of size less than or equal to this value, we
 *  will restart with a higher level of propagation. For problems
 *  of size greater than this value, this higher level of propagation
 *  is too expensive.
 */
#define MAX_RESTART_SIZE       300


/* =================================================================
 *  Routines for creating and destroying a CSP with n variables.
 */
CSP_type *createCSP( int n );
void	 destroyCSP( CSP_type *C );

/* =================================================================
 *  Routines for manipulating the bounds of the domains of the
 *  variables.
 */
void	setBounds( CSP_type *C );
void	incrementUpperBounds( CSP_type *C );
void	saveBounds( CSP_type *C, int currentLevel );
void	restoreBounds( CSP_type *C, int currentLevel );

/* =================================================================
 *  Routine for setting the constraints over the variables.
 */
void	setConstraints( CSP_type *C, DAG_type *DAG );

/* =================================================================
 *  Routines for propagating constraints.
 */
BOOL	doPredsOfNode( CSP_type *C, int v, BOOL pushFlag, int type, int width );
BOOL	doSuccsOfNode( CSP_type *C, int v, BOOL pushFlag, int type, int width );
BOOL	propagate( CSP_type *C, int currentVar, int currentLevel );
void    safePrune( CSP_type *C );

/* =================================================================
 *  Routines for printing/debugging/viewing.
 */
void	printCSP( CSP_type *C );
void	printSolution( CSP_type *C );
void	printGraph( CSP_type *C );
void	verifySolution( CSP_type *C );

/* =================================================================
 *  Routines for timing and restarts.
 */
void	set_start_time( CSP_type *C );
double	elapsed_time( CSP_type *C );
void	set_absolute_CPU_limit( CSP_type *C, double time_limit );
void	set_relative_CPU_limit( CSP_type *C, double time_limit );
BOOL	limit_exceeded( CSP_type *C );

