
#include "global.h"
#include "processor.h"
#include "list.h"
#include "dag.h"

/*
 *  Read in and store the DAG.
 */
DAG_type *
getDAG( char *problem )
{
    DAG_type	*DAG;
    int		i, v_i, v_j, latency, n;
    int		*nPred;
    BOOL	sinkNeeded;
    char	name[256];
    char        line[MAX_LINE_LENGTH];
    FILE	*fp;

    sprintf( name, "C:/Benchmarks/BasicBlocks/%s", problem );
    if( (fp = fopen( name, "r" )) == NULL ) {
	fprintf( stderr, "Could not open file %s\n", name );
	exit( 6 );
    }

    DAG = (DAG_type *)calloc( (size_t)1, sizeof(DAG_type) );

    fgets( line, MAX_LINE_LENGTH, fp );   // skip: ID:benchmark:file:block
    fscanf( fp, "NN:%d\n", &n);		  // read: NN:nodes

    DAG->nInstructions = n+1;		  // add one for dummy root at 0
    DAG->blockID = problem;
    					  // +1 as may need to add a dummy sink node.
    DAG->instructionType = (int *)calloc( (size_t)DAG->nInstructions+1, sizeof(int) );
    DAG->successors = (ELEM **)calloc( (size_t)DAG->nInstructions+1, sizeof(ELEM *) );
    nPred = (int *)calloc( (size_t)DAG->nInstructions+1, sizeof(int) );
    for( i = 0; i <= DAG->nInstructions; i++ ) {
	DAG->successors[i] = (ELEM *)NULL;
	nPred[i] = 0;
    }

    fgets( line, MAX_LINE_LENGTH, fp ); /* read: Node ... */

START:
	/*
	 *  Node ...
	 */
	i = 0;
	while( line[i] != ' ' ) {
	    i++;
	}
	v_i = atoi( &line[i+1] );
	if( (v_i < 0) || (v_i >= DAG->nInstructions) ) {
	    fprintf( stderr, "v_i out of range: %d > %d.\n",
		v_i, DAG->nInstructions );
	    exit( 6 );
	}

	/*
	 *  Assembler instruction ...
	 */
	fgets( line, MAX_LINE_LENGTH, fp );

	/*
	 *  Instruction type ...
	 */
	fgets( line, MAX_LINE_LENGTH, fp );
	i = 0;
	while( line[i] != '\n' ) {
	    i++;
	}
	line[i] = '\0';
	DAG->instructionType[v_i] = getInstructionType( line );

NEXT_EDGE:
	fgets( line, MAX_LINE_LENGTH, fp ); // read: END or Node or s:l */
	if( line[0] == 'E' &&
	    line[1] == 'N' &&
	    line[2] == 'D' ) {
	    goto END;
	}
	if( line[0] == 'N' &&
	    line[1] == 'o' &&
	    line[2] == 'd' &&
	    line[3] == 'e' ) {
	    goto START;
	}

	/*
	 *  Process successor : latency
	 */
	v_j = atoi( &line[0] );
	if( (v_j < 0) || (v_j >= DAG->nInstructions) ) {
	    fprintf( stderr, "v_j out of range: %d > %d.\n",
		v_j, DAG->nInstructions );
	    exit( 6 );
	}
	if( v_i >= v_j ) {
	    fprintf( stderr, "Violates assumption that all successors\n" );
	    fprintf( stderr, "of a vertex have a higher vertex number.\n" );
	    fprintf( stderr, "%d >= %d\n", v_i, v_j );
	    exit( 6 );
	}

	i = 0;
	while( line[i] != ':' ) {
	    i++;
	}
	latency = atoi( &line[i+1] );
	if( (latency < 0) || (latency > MAX_LATENCY) ) {
	    fprintf( stderr, "latency out of range: %d\n", latency );
	    exit( 6 );
	}

	nPred[v_j]++;
	insertList( &(DAG->successors[v_i]), v_j, latency );

	goto NEXT_EDGE;

END:
    /*
     *  Ensure there is a single source node by adding
     *  a dummy node 0 and adding an edge with latency
     *  1 between node 0 and any node without predecessors.
     */
    DAG->instructionType[0] = PSEUDO_INSTR;
    for( i = 1; i < DAG->nInstructions; i++ ) {
	if( nPred[i] == 0 ) { // has no predecessors
	    insertList( &(DAG->successors[0]), i, 1 );
	}
    }
    /*
     *  Ensure there is a single sink node by, if
     *  necessary, adding a dummy node.
     */
    sinkNeeded = FALSE;
    for( i = 1; i < DAG->nInstructions-1; i++ ) {
	if( DAG->successors[i] == NULL ) { // has no successors
	    sinkNeeded = TRUE;
	    break;
	}
    }
    if( sinkNeeded ) {
	DAG->instructionType[DAG->nInstructions] = PSEUDO_INSTR;
	for( i = 1; i < DAG->nInstructions; i++ ) {
	    if( DAG->successors[i] == NULL ) { // has no successors
		insertList( &(DAG->successors[i]), DAG->nInstructions, 1 );
	    }
	}
	DAG->nInstructions++;
    }

    fclose( fp );
    free( nPred );

    return( DAG );
}

/*
 *  Print the DAG for debugging purposes.
 */
void
printDAG( DAG_type *DAG )
{
    int	 i;
    ELEM *q;

    printf( "\nvertices: %d\n", DAG->nInstructions );

    printf( "Successors:\n" );
    for( i = 0; i < DAG->nInstructions; i++ ) {
	printf( "%d", i );
	for( q = DAG->successors[i]; q != NULL; q = q->next ) {
	    printf( " %d:%d", q->vertex, q->latency );
	}
	printf( "\n" );
    }
    printf( "\n" );
}

