
#define DEBUG_REGIONS 0
#define DEBUG_INSERTS 0

#include "global.h"
#include "processor.h"
#include "list.h"
#include "dag.h"
#include "csp.h"

static void
findComponent(
	CSP_type *C,
	int v,
	int label,
	BOOL *inRegion,
	int *component )
{
    int   w;
    ELEM  *p;

    /*
     *  Mark the node v as being in the component.
     */
    component[v] = label;

    /*
     *  For each successor w of v in the region
     *  that is not already part of a component ...
     */
    for( p = C->latencySucc[v]; p != NULL; p = p->next ) {
	w = p->vertex;
	if( !component[w] && inRegion[w] ) {
	    findComponent( C, w, label, inRegion, component );
	}
    }
    /*
     *  For each predecessor w of v in the region
     *  that is not already part of a component ...
     */
    for( p = C->latencyPred[v]; p != NULL; p = p->next ) {
	w = p->vertex;
	if( !component[w] && inRegion[w] ) {
	    findComponent( C, w, label, inRegion, component );
	}
    }

}

static void
fillInNodes(
	int s,
	int t,
	int component[],
	int label,
	int *n,
	int nodes[],
	int nC )
{
    int v;

    *n = 0;
    for( v = 1; v <= nC; v++ ) if( v != s && v != t ) {
	if( component[v] == label ) {
	    nodes[*n] = v;
	    (*n)++;
	}
    }
}

/*
 *  Sort nodes in a component in descending order by critical
 *  path distance (primary key), descending order by number of
 *  successors (secondary key), and increasing order by
 *  instruction type (tertiary key).
 */
static void
sortComponents( CSP_type *C, int t, int nodes[], int n )
{
    int i, current;
    BOOL sorted;
    int temp;

    current = n-1;
    sorted = FALSE;
    while( !sorted ) {
	sorted = TRUE;
	for( i = 0; i < current; i++ ) {
	    if(  (C->cp[nodes[i]][t]   <  C->cp[nodes[i+1]][t]) ||
	        ((C->cp[nodes[i]][t]   == C->cp[nodes[i+1]][t]) &&
		 (C->nSuccs[nodes[i]]  <  C->nSuccs[nodes[i+1]]))   ||
	        ((C->cp[nodes[i]][t]   == C->cp[nodes[i+1]][t]) &&
		 (C->nSuccs[nodes[i]]  == C->nSuccs[nodes[i+1]])    &&
		 (C->varType[nodes[i]] >  C->varType[nodes[i+1]])) ) {
		temp = nodes[i];
		nodes[i] = nodes[i+1];
		nodes[i+1] = temp;
		sorted = FALSE;
	    }
	}
	current--;
    }
}

#if DEBUG_REGIONS
static void
printOneComponent( CSP_type *C, int nodes[], int n, char label )
{
    int   k;
    int   type, nOpsComponent[N_INSTRUCTION_TYPES];

    printf( "\t\tComponent %c: ", label );
    for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
	nOpsComponent[type] = 0;
    }
    for( k = 0; k < n; k++ ) {
	nOpsComponent[C->varType[nodes[k]]]++;
    }
    printf( "[" );
    for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
	printf( "%d ", nOpsComponent[type] );
    }
    printf( "]" );
    for( k = 0; k < n; k++ ) {
	printf( " %d", nodes[k]-1 );
    }
    printf( "\n" );
}
#endif

/*
 *  Return true if components in nodesA and nodesB satisfy theorem.
 *  Sometimes returns FALSE, although with more work it could return TRUE.
 */
static BOOL
checkComponents(
	CSP_type *C,
	int s,		 // NOT USED??? source of the region
	int t,		 // NOT USED??? sink of the region
	int nodesA[],	 // nodesA[0..n-1] contains component A
	int nodesB[],	 // nodesB[0..n-1] contains component B
	int n )
{
    int   i, j, k;
    int   *mapAtoB, *mapBtoA;
    BOOL  *inA, *inB;
    BOOL  conditionsSatisfied;
    ELEM  *p;

    mapAtoB = (int *)calloc( (size_t)C->n+1, sizeof(int) );
    mapBtoA = (int *)calloc( (size_t)C->n+1, sizeof(int) );
    inA = (BOOL *)calloc( (size_t)C->n+1, sizeof(BOOL) );
    inB = (BOOL *)calloc( (size_t)C->n+1, sizeof(BOOL) );

    /*
     *  Order the nodes in the components to correspond to the
     *  1-to-1 mapping, then fill in the mapping.
     *  It may be the case that the components are not isomorphic under
     *  this mapping, yet there exists a mapping under which they are
     *  isomorphic. Thus the routine may return FALSE, although the
     *  correct answer is TRUE.
     */
    //sortComponents( C, t, nodesA, n );
    //sortComponents( C, t, nodesB, n );
    for( k = 0; k < n; k++ ) {
	mapBtoA[nodesB[k]] = nodesA[k];
	mapAtoB[nodesA[k]] = nodesB[k];
	inA[nodesA[k]] = TRUE;
	inB[nodesB[k]] = TRUE;
    }

    conditionsSatisfied = TRUE;

    /*
     *  Check if components A and B are isomorphic.
     *  For each node in the components,
     *  compare the instruction types.
     *  Return false if they differ.
     */
    for( k = 0; k < n; k++ ) {
	if( C->varType[nodesA[k]] != C->varType[nodesB[k]] ) {
	    conditionsSatisfied = FALSE;
#if DEBUG_REGIONS
printf("(a) failure\n");
#endif
	    goto cleanup;
	}
    }

    /*
     *  Check if components A and B are isomorphic.
     *  For each edge entirely within a component,
     *  check that there exists a corresponding edge
     *  in the other component with the same latency.
     *  Return false if they differ.
     */
    for( k = 0; k < n; k++ ) {
	for( p = C->latencySucc[nodesA[k]]; p != NULL; p = p->next ) {
	    //if( (p->vertex != t) && inRegion[p->vertex] &&
	    if( inA[p->vertex] &&
	        (C->cp[mapAtoB[nodesA[k]]][mapAtoB[p->vertex]] != p->latency) ) {
		conditionsSatisfied = FALSE;
#if DEBUG_REGIONS
printf("(b) failure\n");
#endif
		goto cleanup;
	    }
	}
	for( p = C->latencySucc[nodesB[k]]; p != NULL; p = p->next ) {
	    //if( (p->vertex != t) && inRegion[p->vertex] &&
	    if( inB[p->vertex] &&
	        (C->cp[mapBtoA[nodesB[k]]][mapBtoA[p->vertex]] != p->latency) ) {
		conditionsSatisfied = FALSE;
#if DEBUG_REGIONS
printf("(c) failure\n");
#endif
		goto cleanup;
	    }
	}
    }

    /*
     *  Condition 0 in Theorem.
     *  Ensure independence of corresponding nodes by
     *  enforcing something stronger: ensure that there
     *  are no edges between components A and B.
     */
    for( i = 0; i < n; i++ )
    for( j = 0; j < n; j++ ) {
	if( C->cp[nodesA[i]][nodesB[j]] >= 0 ||
	    C->cp[nodesB[j]][nodesA[i]] >= 0 ) {
	    conditionsSatisfied = FALSE;
#if DEBUG_REGIONS
printf("(d) failure: not independent\n");
#endif
	    goto cleanup;
	}
    }

    /*
     *  Condition 1 in Theorem.
     *  For each edge (v,w) such that v not in A and w in component A,
     *  compare the latency on the edge to the critical path distance from
     *  v to w's corresponding node in component B.
     */
    for( k = 0; k < n; k++ ) {
	for( p = C->latencyPred[nodesA[k]]; p != NULL; p = p->next ) {
	    if( !inA[p->vertex] ) {
		if( p->latency > C->cp[p->vertex][mapAtoB[nodesA[k]]] ) {
		    conditionsSatisfied = FALSE;
#if DEBUG_REGIONS
printf("(e) failure: (%d,%d):%d > (%d,%d):%d\n",
p->vertex-1,nodesA[k]-1,p->latency,
p->vertex-1,mapAtoB[nodesA[k]]-1,C->cp[p->vertex][mapAtoB[nodesA[k]]]);
#endif
		    goto cleanup;
		}
	    }
	}
    }

    /*
     *  Condition 2 in Theorem.
     *  For each edge (v,w) such that v in component B and w not in B,
     *  compare the latency on the edge to the critical path distance from
     *  v's corresponding node in component A to w.
     */
    for( k = 0; k < n; k++ ) {
	for( p = C->latencySucc[nodesB[k]]; p != NULL; p = p->next ) {
	    if( !inB[p->vertex] ) {
		if( p->latency > C->cp[mapBtoA[nodesB[k]]][p->vertex] ) {
		    conditionsSatisfied = FALSE;
#if DEBUG_REGIONS
printf("(f) failure: (%d,%d):%d > (%d,%d):%d\n",
nodesB[k]-1,p->vertex-1,p->latency,
mapBtoA[nodesB[k]]-1,p->vertex-1,C->cp[mapBtoA[nodesB[k]]][p->vertex]);
#endif
		    goto cleanup;
		}
	    }
	}
    }

cleanup:
    free( mapAtoB );
    free( mapBtoA );
    free( inA );
    free( inB );

    return( conditionsSatisfied );
}

int
setDominanceConstraints2( CSP_type *C, int s, int t, BOOL *inRegion, char ch )
{
    int   k, v, w, nComponents, latency, n_min, nConstraints;
    int   *component;
    int   i, n_i, *nodes_i;
    int   j, n_j, *nodes_j;
    BOOL  conditionsSatisfied;
    ELEM  *p;

    component = (int *)calloc( (size_t)C->n+1, sizeof(int) );
    nodes_i   = (int *)calloc( (size_t)C->n+1, sizeof(int) );
    nodes_j   = (int *)calloc( (size_t)C->n+1, sizeof(int) );

    /*
     *  Find the components in the region.
     *  For each successor w of s in the interior of the region
     *  that is not already part of a component ...
     */
    nComponents = 0;
    for( p = C->latencySucc[s]; p != NULL; p = p->next ) {
	w = p->vertex;
// IMPROVEMENT: Should test that t is reachable from w, skip otherwise
	if( inRegion[w] && !component[w] ) {
	    nComponents++;
	    component[s] = nComponents;
	    component[t] = nComponents;
	    findComponent( C, w, nComponents, inRegion, component );
	}
    }

    /*
     *  Find pairs of components that satisfy theorem.
     *  This can be improved by passing nodeCount into the routine
     *  and skipping these tests if a component is larger than
     *  nodeCount/2.
     */
    nConstraints = 0;
    for( i =   1; i <= nComponents; i++ )
    for( j = i+1; j <= nComponents; j++ ) {
	fillInNodes( s, t, component, i, &n_i, nodes_i, C->n );
	fillInNodes( s, t, component, j, &n_j, nodes_j, C->n );
#if DEBUG_REGIONS
	printf( "Region (%d, %d):\n", s-1, t-1 );
	printf( "\tConsidering components:\n" );
	printOneComponent( C, nodes_i, n_i, 'i' );
	printOneComponent( C, nodes_j, n_j, 'j' );
#endif
	if( n_i < n_j ) n_min = n_i;
	else		n_min = n_j;

	/*
	 *  Test if the components satisfy the theorem or can
	 *  be made to satisfy the theorem by dropping extra
	 *  nodes from the tail.
	 */
	conditionsSatisfied = checkComponents( C, s, t,
					nodes_i, nodes_j, n_min );

	/*
	 *  Test if the components can be made to satisfy the
	 *  theorem by dropping extra nodes from the head.
	 */
	if( !conditionsSatisfied && (n_i != n_j) ) {
	    for( k = 0; k < n_min; k++ ) {
		if( n_i > n_j ) nodes_i[k] = nodes_i[k + (n_i - n_j)];
		else		nodes_j[k] = nodes_j[k + (n_j - n_i)];
	    }
	    conditionsSatisfied = checkComponents( C, s, t,
					nodes_i, nodes_j, n_min );
	}

if( !conditionsSatisfied ) {
	fillInNodes( s, t, component, j, &n_i, nodes_i, C->n );
	fillInNodes( s, t, component, i, &n_j, nodes_j, C->n );

	    if( n_i < n_j ) n_min = n_i;
	    else	    n_min = n_j;

	    /*
	     *  Test if the components satisfy the theorem or can
	     *  be made to satisfy the theorem by dropping extra
	     *  nodes from the tail.
	     */
	    conditionsSatisfied = checkComponents( C, s, t,
				nodes_i, nodes_j, n_min );

	    /*
	     *  Test if the components can be made to satisfy the
	     *  theorem by dropping extra nodes from the head.
	     */
	    if( !conditionsSatisfied && (n_i != n_j) ) {
	    //if( !conditionsSatisfied && ((n_i == n_j+1) || (n_i+1 == n_j)) ) {
		for( k = 0; k < n_min; k++ ) {
		    //if( n_i > n_j ) nodes_i[k] = nodes_i[k+1];
		    //else	    nodes_j[k] = nodes_j[k+1];
		    if( n_i > n_j ) nodes_i[k] = nodes_i[k + (n_i - n_j)];
		    else	    nodes_j[k] = nodes_j[k + (n_j - n_i)];
		}
		conditionsSatisfied = checkComponents( C, s, t,
					nodes_i, nodes_j, n_min );
/*
*/

	    }
}

	    /*
	     *  If conditions of theorem are satisfied, insert a
	     *  dominance constraint between each matching pair of
	     *  nodes in the components.
	     *
	     *  The same constraints can be discovered more than once.
	     *  The test ensures that the same edge is not inserted
	     *  more than once and that no simple cycles are introduced.
	     *  Updating the cp distance marks the edge as inserted.
	     *  Must also ensure that no longer cycles are introduced
	     *  by being careful how this routine is called.
	     */
	    if( conditionsSatisfied ) {
		for( k = 0; k < n_min; k++ ) {
nConstraints++;
		    v = nodes_i[k];
		    w = nodes_j[k];
		    if( nFunctionalUnits( C->varType[v] ) == 1 ) latency = 1;
		    else					 latency = 0;

		    if( !C->edgeMarked[v][w] && !C->edgeMarked[w][v] ) {
		    //if( (latency > C->cp[v][w]) && !C->edgeMarked[w][v] ) {

#if DEBUG_INSERTS
		        printf("%c: insert dominance constraint: %d --%d--> %d\n",ch,v-1,latency,w-1);
#endif
		        insertList( &C->dominanceSucc[v], w, latency );
		        insertList( &C->dominancePred[w], v, latency );
		        C->edgeMarked[v][w] = TRUE;
		    }
		}
	    }
	//}
    }

    free( component );
    free( nodes_i );
    free( nodes_j );

    return( nConstraints );
}

static void
sortmin( node_type v[], int n )
{
    int i, current;
    BOOL sorted;
    int t;

    current = n-1;
    sorted = FALSE;
    while( !sorted ) {
	sorted = TRUE;
	for( i = 0; i < current; i++ ) {
	    if( (v[i].bound  > v[i+1].bound) ||
	       ((v[i].bound == v[i+1].bound) &&
		(v[i].bound2 > v[i+1].bound2)) ||
	       ((v[i].bound == v[i+1].bound) &&
	        (v[i].bound2 == v[i+1].bound2) &&
		(v[i].data > v[i+1].data)) ) {
		t = v[i].bound;
		v[i].bound = v[i+1].bound;
		v[i+1].bound = t;
		t = v[i].data;
		v[i].data = v[i+1].data;
		v[i+1].data = t;
		sorted = FALSE;
	    }
	}
	current--;
    }
}

int
setDominanceConstraints1( CSP_type *C )
{
    int   v, w, latency, nConstraints;
    int   i, n_i, *nodes_i;
    int   j, n_j, *nodes_j;
    BOOL  conditionsSatisfied;
    int   type, lower, upper;
    int   count;

    nodes_i   = (int *)calloc( (size_t)C->n+1, sizeof(int) );
    nodes_j   = (int *)calloc( (size_t)C->n+1, sizeof(int) );

    nConstraints = 0;

    for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
        count = 0;
        for( v = 1; v <= C->n; v++ ) {
            if( C->varType[v] == type ) {
                C->values[count].bound = C->cp[1][v];
                C->values[count].bound2 = C->cp[v][C->n];
                C->values[count].data  = v;
                count++;
            }
        }
    
        if( count <= 1 ) continue;
    
        sortmin( C->values, count );
    
        lower = 0;
        upper = 0;
    
        while( upper < count ) {
            lower = upper;
            upper = lower+1;
            while( upper < count &&
                   C->values[upper].bound  == C->values[lower].bound &&
                   C->values[upper].bound2 == C->values[lower].bound2 ) {
                upper++;
            }
    
            for( i = lower; i < upper; i++ )
            for( j = i+1;   j < upper; j++ ) {
    
                nodes_i[0] = C->values[i].data; n_i = 1;
                nodes_j[0] = C->values[j].data; n_j = 1;
    
                conditionsSatisfied = checkComponents( C, i, j,
                                            nodes_i, nodes_j, 1 );
                if( conditionsSatisfied ) {
                    nConstraints++;
                    v = nodes_i[0];
                    w = nodes_j[0];
                    if( nFunctionalUnits( C->varType[v] ) == 1 ) latency = 1;
                    else                                         latency = 0;
    
                    if( !C->edgeMarked[v][w] && !C->edgeMarked[w][v] ) {
    #if DEBUG_INSERTS
    printf("Pass 1: insert dominance constraint: %d --%d--> %d\n",v-1,latency,w-1);
    #endif
                        insertList( &C->dominanceSucc[v], w, latency );
                        insertList( &C->dominancePred[w], v, latency );
                        C->edgeMarked[v][w] = TRUE;
                    }
                    break;
                }
            }
        }
    }

    free( nodes_i );
    free( nodes_j );

    return( nConstraints );
}

