
#include "global.h"
#include "processor.h"
#include "list.h"
#include "dag.h"
#include "ls.h"

/* =================================================================
 *  Routines for creating and destroying an LS with n variables.
 */
static LS_type  *createLS( int n );
static void	destroyLS( LS_type *L );

/* =================================================================
 *  Routines for finding the best instruction to schedule.
 */
static int	findBestInstruction( LS_type *L, int type, int currentTime );
static void	setLatencyConstraints( LS_type *L, DAG_type *DAG );
static void	setEST( LS_type *L );
static void	updateEST( LS_type *L, int v, int issueTime );
static void	updatePreds( LS_type *L, int v );
static void	setMaxDelay( LS_type *L );

/* =================================================================
 *  Routines for printing/debugging.
 */
static void	printLS( LS_type *L );
static void	verifySolution( LS_type *L, int order[] );

int
listScheduler( DAG_type *DAG, int order[] )
{
    int      v, type, length;
    int	     bestInstruction, room[N_INSTRUCTION_TYPES];
    int	     instructionsToSchedule, currentTime;
    BOOL     anotherPassNeeded;
    LS_type  *L;

    L = createLS( DAG->nInstructions );
    for( v = 1; v <= L->n; v++ ) {
	L->varType[v] = DAG->instructionType[v-1];
	L->isScheduled[v] = FALSE;
    }

    setLatencyConstraints( L, DAG );
    setEST( L );
    setMaxDelay( L );

    currentTime = 0;
    instructionsToSchedule = L->n;
    while( instructionsToSchedule > 0 ) {
	currentTime++;

	/*
	 *  Schedule the best instructions available at the
	 *  current time cycle.
	 */
	for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
	    for( room[type]=nFunctionalUnits( type ); room[type] > 0; room[type]-- ) {
		bestInstruction = findBestInstruction( L, type, currentTime );
		if( bestInstruction == 0 ) break;
		order[bestInstruction-1] = currentTime;
		L->isScheduled[bestInstruction] = TRUE;
		instructionsToSchedule--;
		updatePreds( L, bestInstruction );
		updateEST( L, bestInstruction, currentTime );
	    }
	}

	/*
	 *  Check whether any new instructions are available
	 *  to be scheduled as a result of the previous pass.
	 */
	do {
	    anotherPassNeeded = FALSE;
	    for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
		for( ; room[type] > 0; room[type]-- ) {
		    bestInstruction = findBestInstruction( L, type, currentTime );
		    if( bestInstruction == 0 ) break;
		    anotherPassNeeded = TRUE;
		    order[bestInstruction-1] = currentTime;
		    L->isScheduled[bestInstruction] = TRUE;
		    instructionsToSchedule--;
		    updatePreds( L, bestInstruction );
		    updateEST( L, bestInstruction, currentTime );
		}
	    }
	} while( anotherPassNeeded );
    }

    verifySolution( L, order );
    length = order[L->n-1];

    destroyLS( L );

    return( length );
}

/* =================================================================
 *  Routines for creating and destroying an LS with n variables.
 */

static LS_type *
createLS( int n )
{
    int       i;
    LS_type   *L;

    L = (LS_type *)calloc( (size_t)1, sizeof(LS_type) );

    L->n = n;

    L->varType     = (int *)calloc( (size_t)n+1, sizeof(int) );
    L->maxDelay    = (int *)calloc( (size_t)n+1, sizeof(int) );
    L->EST         = (int *)calloc( (size_t)n+1, sizeof(int) );
    L->nPreds      = (int *)calloc( (size_t)n+1, sizeof(int) );
    L->nSuccs      = (int *)calloc( (size_t)n+1, sizeof(int) );
    L->isScheduled = (BOOL *)calloc( (size_t)n+1, sizeof(BOOL) );

    L->latencySucc = (ELEM **) calloc( (size_t)n+1, sizeof(ELEM *) );
    L->latencyPred = (ELEM **) calloc( (size_t)n+1, sizeof(ELEM *) );
    for( i = 1; i <= n; i++ ) {
	L->latencySucc[i] = (ELEM *)NULL;
	L->latencyPred[i] = (ELEM *)NULL;
    }

    return( L );
}

static void
destroyLS( LS_type *L )
{
    int  i;

    free( L->varType );
    free( L->maxDelay );
    free( L->EST );
    free( L->isScheduled );
    free( L->nPreds );
    free( L->nSuccs );

    for( i = 1; i <= L->n; i++ ) {
	destroyList( L->latencySucc[i] );
	destroyList( L->latencyPred[i] );
    }
    free( L->latencySucc );
    free( L->latencyPred );

    free( L );
}

/*
 *  Set the latency constraints.
 *
 *  latencySucc[w] is a list of all nodes v and latencies l
 *	such that there is a latency constraint between w and v
 *	of the form: lower[w] + l <= lower[v].
 *  latencyPred[w] is a list of all nodes v and latencies l
 *	such that there is a latency constraint between w and v
 *	of the form: upper[w] - l >= upper[v].
 */
static void
setLatencyConstraints( LS_type *L, DAG_type *DAG )
{
    int  v, w, latency;
    ELEM *q;

    for( v = 0; v < L->n-1; v++ ) {
	for( q = DAG->successors[v]; q != NULL; q = q->next ) {
	    w       = q->vertex;
	    latency = q->latency;	// latency on edge (v,w)
	    /*
	     *  Externally, the vertices are named 0, ..., n-1.
	     *  Internally, the vertices are named 1, ..., n.
	     */
	    insertList( &L->latencySucc[v+1], w+1, latency );
	    insertList( &L->latencyPred[w+1], v+1, latency );
	}
    }

    for( v = 1; v <= L->n; v++ ) {
	L->nPreds[v] = length( L->latencyPred[v] );
	L->nSuccs[v] = length( L->latencySucc[v] );
    }
}

/*
 *  Set the earliest start time of the instructions by
 *  propagating the latency constraints.
 */
static void
setEST( LS_type *L )
{
    int  v, w;
    ELEM *p;

    for( v = 1; v <= L->n; v++ ) {
	L->EST[v] = 1;
    }

    for( v = 1; v <= L->n; v++ ) {
	for( p = L->latencySucc[v]; p != NULL; p = p->next ) {
	    w = p->vertex;
	    if( L->EST[w] < (L->EST[v] + p->latency) ) {
		L->EST[w] =  L->EST[v] + p->latency;
	    }
	}
    }
}

/*
 *  Update the earliest start times of the instructions that
 *  are the successors of a recently issued instruction by
 *  propagating the latency constraints.
 */
static void
updateEST( LS_type *L, int v, int issueTime )
{
    int  w;
    ELEM *p;

    for( p = L->latencySucc[v]; p != NULL; p = p->next ) {
	w = p->vertex;
	if( L->EST[w] < (issueTime + p->latency) ) {
	    L->EST[w] =  issueTime + p->latency;
	}
    }
}

/*
 *  Update the ready list by checking whether the successors
 *  of a recently issued instruction are now ready to be
 *  scheduled.
 */
static void
updatePreds( LS_type *L, int v )
{
    int  w;
    ELEM *p;

    for( p = L->latencySucc[v]; p != NULL; p = p->next ) {
	w = p->vertex;
	L->nPreds[w]--;
    }
}

/*
 *  Set the maxDelay by propagating the latency constraints.
 */
static void
setMaxDelay( LS_type *L )
{
    int  v, w;
    ELEM *p;

    for( v = L->n; v >= 1; v-- ) {
	L->maxDelay[v] = 1;
    }

    for( v = L->n; v >= 1; v-- ) {
	for( p = L->latencyPred[v]; p != NULL; p = p->next ) {
	    w = p->vertex;
	    if( L->maxDelay[w] < (L->maxDelay[v] + p->latency) ) {
		L->maxDelay[w] =  L->maxDelay[v] + p->latency;
	    }
	}
    }
}

/*
 *  Find instruction with max delay.
 *  Break ties by lowest EST.
 */
static int
findBestInstruction( LS_type *L, int type, int currentTime  )
{
    int v;
    int	bestInstruction, maxDelay;

    bestInstruction = 0;
    maxDelay = 0;
    for( v = 1; v <= L->n; v++ ) {
	/*
	 *  An instruction v is ready to be scheduled if all
	 *  of its predecessors are scheduled (nPreds[v] == 0)
	 *  and enough time has passed to satisfy the latency
	 *  constraints (EST[v] <= currentTime)
	 */
	if((L->nPreds[v] == 0) &&
	   !L->isScheduled[v] &&
	    L->varType[v] == type &&
	    L->EST[v] <= currentTime ) {

	    if( L->maxDelay[v] > maxDelay ) {
		maxDelay = L->maxDelay[v];
		bestInstruction = v;
	    }
	    else
	    if( (L->maxDelay[v] == maxDelay) &&
	        (L->EST[v] < L->EST[bestInstruction]) ) {
		bestInstruction = v;
	    }
	    /*
	    else
	    if( (L->maxDelay[v] == maxDelay) &&
		(L->EST[v] == L->EST[bestInstruction]) &&
		(L->nSuccs[v] > L->nSuccs[bestInstruction]) ) {
		bestInstruction = v;
	    }
	    */
	}
    }

    return( bestInstruction );
}

/* =================================================================
 *  Routines for printing/debugging.
 */

static void
printLS( LS_type *L )
{
    int  v;
    ELEM *p;

    printf( "\n\tn: %d\n", L->n );
    for( v = 1; v <= L->n; v++ ) {
	printf( "\t%3d: est = %3d maxDelay = %3d nPreds = %3d isSched = %d; type=%d\n",
	    v,
	    L->EST[v],
	    L->maxDelay[v],
	    L->nPreds[v],
	    L->isScheduled[v],
	    L->varType[v] );
    }
    printf( "\n" );

    printf( "Successors:\n" );
    for( v = 1; v <= L->n; v++ ) {
	printf( "%d", v );
	for( p = L->latencySucc[v]; p != NULL; p = p->next ) {
	    printf( " %d:%d", p->vertex, p->latency );
	}
	printf( "\n" );
    }
    printf( "\n" );

    printf( "Predecessors:\n" );
    for( v = 1; v <= L->n; v++ ) {
	printf( "%d", v );
	for( p = L->latencyPred[v]; p != NULL; p = p->next ) {
	    printf( " %d:%d", p->vertex, p->latency );
	}
	printf( "\n" );
    }
    printf( "\n" );
}

static void
verifySolution( LS_type *L, int order[] )
{
    int  t, v, m, type, status, *count;
    ELEM *p;

    m = 0;
    for( v = 1; v <= L->n; v++ ) {
	if( m < order[v-1] ) m = order[v-1];
    }
    count = (int *)calloc( (size_t)m+1, sizeof(int) );

    status = 1;
    for( v = 1; v <= L->n; v++ ) {
	status = status && (order[v-1] > 0);
    }
    if( !status ) {
	printf( "Verifying List Scheduler solution:\n" );
	printf( "\tEach instruction issued not in range [1, %d].\n", m );
	exit( 6 );
    }

    for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
	for( t = 1; t <= m; t++ ) {
	    count[t] = 0;
	}
	for( v = 1; v <= L->n; v++ ) {
	    if( L->varType[v] == type )
		count[order[v-1]]++;
	}
	status = 1;
	for( t = 1; t <= m; t++ ) {
	    status = status && (count[t] <= nFunctionalUnits( type ));
	}
	if( !status ) {
	    printf( "Verifying List Scheduler solution:\n" );
	    printf( "\tIssue-width constraint on type %d not satisfied.\n", type );
	    exit( 6 );
	}
    }

    status = 1;
    for( v = 1; v <= L->n; v++ ) {
	for( p = L->latencySucc[v]; p != NULL; p = p->next ) {
	    status = status &&
			((order[v-1] + p->latency) <= order[p->vertex-1]);
	}
    }
    if( !status ) {
	printf( "Verifying List Scheduler solution:\n" );
	printf( "\tSuccessor constraints not satisfied.\n" );
	exit( 6 );
    }

    status = 1;
    for( v = 1; v <= L->n; v++ ) {
	for( p = L->latencyPred[v]; p != NULL; p = p->next ) {
	    status = status &&
			((order[p->vertex-1] + p->latency) <= order[v-1]);
	}
    }
    if( !status ) {
	printf( "Verifying List Scheduler solution:\n" );
	printf( "\tPredecessor constraints not satisfied.\n" );
	exit( 6 );
    }

    free( count );
}

