
#include <math.h>
#include "global.h"
#include "processor.h"
#include "list.h"
#include "dag.h"
#include "csp.h"

/*
 *  Internally defined routines.
 */
static BOOL propagateIncrementalConstraints( CSP_type *C );
static BOOL propagateBinaryGCC( CSP_type *C, int type, int width );
static BOOL propagateGCC(       CSP_type *C, int type, int width );
static BOOL safePruneUpper(     CSP_type *C, int type, int width );
static BOOL safePruneLower(     CSP_type *C, int type, int width );


static void
sortmin( node_type v[], int n )
{
    int i, current;
    BOOL sorted;
    int t;

    current = n-1;
    sorted = FALSE;
    while( !sorted ) {
	sorted = TRUE;
	for( i = 0; i < current; i++ ) {
	    if( v[i].bound > v[i+1].bound ) {
		t = v[i].bound;
		v[i].bound = v[i+1].bound;
		v[i+1].bound = t;
		t = v[i].data;
		v[i].data = v[i+1].data;
		v[i+1].data = t;
		sorted = FALSE;
	    }
	}
	current--;
    }
}

int compare2(const void *x, const void *y)
{
    return( (*((node_type *)x)).bound - (*((node_type *)y)).bound );
}

/*
 *  Assumes that a vertex appears at most once on an adjacency list.
 */
BOOL
doPredsOfNode( CSP_type *C, int v, BOOL pushFlag, int type, int width )
{
    int  i, j, count, lb;
    BOOL consistent;
    int  minLatency;	      // minimum latency among remaining predecessors
    int  minLatencyLocation;  // location of minimum latency in array
    ELEM *p;

    if( C->nPreds[v] <= width ) {
	return( TRUE );
    }

    /*
     *  The amount of propagation can be increased (at extra cost) by
     *  adding all predecessors with this type of operation; i.e.,
     *  for( i = v-1; i > 1; i-- ) {
     *      if( (C->varType[i] == type) && (C->cp[i][v] >= 0) ) {
     *  The above code is only valid if the critical path distances have
     *  been set prior to calling this routine.
     *  Also, this assumes that i < j whenever there is an arc
     *  from i to j. This may not be true after dominance constraints
     *  have been added.
     */
    if( !pushFlag ) {
        count = 0;
        for( i = 1; i < v; i++ ) if( C->cp[i][v] >= 0 ) {
    	    if( C->varType[i] == type ) {
    	        C->values[count].bound = C->lower[i];
    	        C->values[count].data  = C->cp[i][v];
    	        count++;
    	    }
        }
    }
    else {
        count = 0;
        for( p = C->latencyPred[v]; p != NULL; p = p->next ) {
    	    if( C->varType[p->vertex] == type ) {
    	        C->values[count].bound = C->lower[p->vertex];
    	        C->values[count].data  = p->latency;
    	        count++;
    	    }
        }
    }

    if( count <= width ) {
	return( TRUE );
    }

    qsort(C->values, count, sizeof(node_type), compare2);
    //sortmin( C->values, count );

    /*
     *  The constraint can only have an effect if the number of predecessors
     *  with the same lower bound is greater than the width.
     */
    i = 0;
    while( (i < count) && (C->values[i].bound != C->values[i+1].bound) ) {
	i++;
    }
    minLatencyLocation = -HUGE;
    minLatency = HUGE;
    for( ; i < count; i++ ) {
	/*
	 *  If the location of the minLatency has been stepped past,
	 *  find a new minLatency and record the location.
	 */
	if( minLatencyLocation < i ) {
	    minLatency = HUGE;
	    for( j = i; j < count; j++ ) {
		if( minLatency >= C->values[j].data ) {
		    minLatency = C->values[j].data;
		    minLatencyLocation = j;
		}
	    }
	}

	lb =  C->values[i].bound
	    + (int)ceil(((double)count-i)/((double)width))
	    - 1
	    + minLatency;

	if( C->lower[v] < lb ) {
	    C->lower[v] = lb;
	    if( C->lower[v] > C->upper[v] ) {
		return( FALSE );
	    }
	    if( pushFlag ) {
		push( v, C->s_lower );
		C->changes++;
		C->iterNeeded[C->varType[v]] = 1;
		if( C->lower[v] == C->upper[v] ) {
		    consistent =
			propagateBinaryGCC( C, type, width );
		    if( !consistent ) return( FALSE );
		}
	    }
	}
    }

    return( TRUE );
}

static void
sortmax( node_type v[], int n )
{
    int i, current;
    BOOL sorted;
    int t;

    current = n-1;
    sorted = FALSE;
    while( !sorted ) {
	sorted = TRUE;
	for( i = 0; i < current; i++ ) {
	    if( v[i].bound < v[i+1].bound ) {
		t = v[i].bound;
		v[i].bound = v[i+1].bound;
		v[i+1].bound = t;
		t = v[i].data;
		v[i].data = v[i+1].data;
		v[i+1].data = t;
		sorted = FALSE;
	    }
	}
	current--;
    }
}

int compare(const void *x, const void *y)
{
    return( (*((node_type *)y)).bound - (*((node_type *)x)).bound );
}

/*
 *  Assumes that a vertex appears at most once on an adjacency list.
 */
BOOL
doSuccsOfNode( CSP_type *C, int v, BOOL pushFlag, int type, int width )
{
    int  i, j, count, ub;
    BOOL consistent;
    int  minLatency;	      // minimum latency among remaining predecessors
    int  minLatencyLocation;  // location of minimum latency in array
    ELEM *p;

    if( C->nSuccs[v] <= width ) {
	return( TRUE );
    }

    /*
     *  The amount of propagation can be increased (at extra cost) by
     *  adding all predecessors with this type of operation; i.e.,
     *  for( i = v+1; i < C->n; i++ ) {
     *      if( (C->varType[i] == type) && (C->cp[v][i] >= 0) ) {
     *  The above code is only valid if the critical path distances have
     *  been set prior to calling this routine.
     *  Also, this assumes that i < j whenever there is an arc
     *  from i to j. This may not be true after dominance constraints
     *  have been added.
     */
    if( !pushFlag ) {
        count = 0;
        for( i = v+1; i <= C->n; i++ ) if( C->cp[v][i] >= 0 ) {
            if( C->varType[i] == type ) {
                C->values[count].bound = C->upper[i];
                C->values[count].data  = C->cp[v][i];
                count++;
            }
      }
    }
    else {
        count = 0;
        for( p = C->latencySucc[v]; p != NULL; p = p->next ) {
            if( C->varType[p->vertex] == type ) {
                C->values[count].bound = C->upper[p->vertex];
                C->values[count].data  = p->latency;
                count++;
            }
        }
    }

    if( count <= width ) {
	return( TRUE );
    }

    qsort(C->values, count, sizeof(node_type), compare);
    //sortmax( C->values, count );

    /*
     *  The constraint can only have an effect if the number of predecessors
     *  with the same lower bound is greater than the width.
     */
    i = 0;
    while( (i < count) && (C->values[i].bound != C->values[i+1].bound) ) {
	i++;
    }
    minLatencyLocation = -HUGE;
    minLatency = HUGE;
    for( ; i < count; i++ ) {
	/*
	 *  If the location of the minLatency has been stepped past,
	 *  find a new minLatency and record the location.
	 */
	if( minLatencyLocation < i ) {
	    minLatency = HUGE;
            for( j = i; j < count; j++ ) {
	        if( minLatency >= C->values[j].data ) {
		    minLatency =  C->values[j].data;
	        }
	    }
	}

	ub =  C->values[i].bound
	    - (int)ceil(((double)count-i)/((double)width))
	    + 1
	    - minLatency;

	if( C->upper[v] > ub ) {
	    C->upper[v] = ub;
	    if( C->lower[v] > C->upper[v] ) {
		return( FALSE );
	    }
	    if( pushFlag ) {
		push( v, C->s_upper );
		C->changes++;
		C->iterNeeded[C->varType[v]] = 1;
		if( C->lower[v] == C->upper[v] ) {
		    consistent =
			propagateBinaryGCC( C, type, width );
		    if( !consistent ) return( FALSE );
		}
	    }
	}
    }

    return( TRUE );
}

static void
cleanupStacks( CSP_type *C )
{
    int  v;

    while( !stackEmpty( C->s_lower ) ) { v = pop( C->s_lower ); }
    while( !stackEmpty( C->s_upper ) ) { v = pop( C->s_upper ); }
}

BOOL
propagate( CSP_type *C, int currentVar, int currentLevel )
{
    int  type;
    BOOL consistent;

    cleanupStacks( C );

    C->changes = 0;

    push( currentVar, C->s_lower );
    push( currentVar, C->s_upper );

    type = C->varType[currentVar];
    consistent =
	propagateBinaryGCC( C, type, nFunctionalUnits( type ) );
    if( !consistent ) return( FALSE );

    while( !stackEmpty( C->s_lower ) || !stackEmpty( C->s_upper ) ) {

	for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
	    if( currentLevel == 0 ) C->iterNeeded[type] = 1;
	    else		    C->iterNeeded[type] = 0;
	}
	C->iterNeeded[C->varType[currentVar]] = 1;

	while( !stackEmpty( C->s_lower ) || !stackEmpty( C->s_upper ) ) {
	    consistent = propagateIncrementalConstraints( C );
	    if( !consistent ) return( FALSE );
	}

	for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
	    if( C->iterNeeded[type] && C->nOps[type] > nFunctionalUnits( type ) ) {
		consistent =
		    propagateGCC( C, type, nFunctionalUnits( type ) );
		if( !consistent ) return( FALSE );
	    }
	}

    }

    return( TRUE );
}

static BOOL
propagateIncrementalConstraints( CSP_type *C )
{
    int  v, w, type;
    BOOL consistent;
    ELEM *p;

    /*
     *  Propagate lower bound changes.
     */
    while( !stackEmpty( C->s_lower ) ) {
	v = pop( C->s_lower );
	/*
	 *  for each constraint involving v...
	 */
	for( p = C->latencySucc[v]; p != NULL; p = p->next ) {
	    w = p->vertex;
	    if( C->lower[w] < (C->lower[v] + p->latency) ) {
		C->lower[w] =  C->lower[v] + p->latency;
		if( C->lower[w] > C->upper[w] ) {
		    return( FALSE );
		}
		push( w, C->s_lower );
		C->changes++;
		C->iterNeeded[C->varType[w]] = 1;
		if( C->lower[w] == C->upper[w] ) {
		    type = C->varType[w];
		    consistent =
			propagateBinaryGCC( C, type, nFunctionalUnits( type ) );
		    if( !consistent ) return( FALSE );
		}
	    }
	}

	for( p = C->distanceSucc[v]; p != NULL; p = p->next ) {
	    w = p->vertex;
	    if( C->lower[w] < (C->lower[v] + p->latency) ) {
		C->lower[w] =  C->lower[v] + p->latency;
		if( C->lower[w] > C->upper[w] ) {
		    return( FALSE );
		}
		push( w, C->s_lower );
		C->changes++;
		C->iterNeeded[C->varType[w]] = 1;
		if( C->lower[w] == C->upper[w] ) {
		    type = C->varType[w];
		    consistent =
			propagateBinaryGCC( C, type, nFunctionalUnits( type ) );
		    if( !consistent ) return( FALSE );
		}
	    }
	}

	for( p = C->edgePred[v]; p != NULL; p = p->next ) {
	    w = p->vertex;
	    if( C->nPreds[w] > 1 ) { // isn't this always true?
		for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
		    consistent =
			doPredsOfNode( C, w, /* pushFlag */TRUE,
						type, nFunctionalUnits( type ) );
		    if( !consistent ) return( FALSE );
		}
	    }
	}
    }

    /*
     *  Propagate upper bound changes.
     */
    while( !stackEmpty( C->s_upper ) ) {
	v = pop( C->s_upper );
	/*
	 *  for each constraint involving v...
	 */
	for( p = C->latencyPred[v]; p != NULL; p = p->next ) {
	    w = p->vertex;
	    if( C->upper[w] > (C->upper[v] - p->latency) ) {
		C->upper[w] =  C->upper[v] - p->latency;
		if( C->lower[w] > C->upper[w] ) {
		    return( FALSE );
		}
		push( w, C->s_upper );
		C->changes++;
		C->iterNeeded[C->varType[w]] = 1;
		if( C->lower[w] == C->upper[w] ) {
		    type = C->varType[w];
		    consistent =
			propagateBinaryGCC( C, type, nFunctionalUnits( type ) );
		    if( !consistent ) return( FALSE );
		}
	    }
	}

	for( p = C->distancePred[v]; p != NULL; p = p->next ) {
	    w = p->vertex;
	    if( C->upper[w] > (C->upper[v] - p->latency) ) {
		C->upper[w] =  C->upper[v] - p->latency;
		if( C->lower[w] > C->upper[w] ) {
		    return( FALSE );
		}
		push( w, C->s_upper );
		C->changes++;
		C->iterNeeded[C->varType[w]] = 1;
		if( C->lower[w] == C->upper[w] ) {
		    type = C->varType[w];
		    consistent =
			propagateBinaryGCC( C, type, nFunctionalUnits( type ) );
		    if( !consistent ) return( FALSE );
		}
	    }
	}

	for( p = C->edgeSucc[v]; p != NULL; p = p->next ) {
	    w = p->vertex;
	    if( C->nSuccs[w] > 1 ) {
		for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
		    consistent =
			doSuccsOfNode( C, w, /* pushFlag */TRUE,
						type, nFunctionalUnits( type ) );
		    if( !consistent ) return( FALSE );
		}
	    }
	}
    }

    return( TRUE );
}

/*
 *  Propagate special case of gcc constraint.
 */
static BOOL
propagateBinaryGCC( CSP_type *C, int type, int width )
{
    int  i, v;

    if( C->nOps[type] <= width ) {
	return( TRUE );
    }

    /*
     *  Shouldn't happen: filled array is too small.
     */
    if( C->m > (C->n*MAX_LATENCY) ) {
	return( TRUE );
    }

    for( i = 1; i <= C->m; i++ ) {
	C->filled[i] = 0;
    }

    for( v = 1; v <= C->n; v++ ) if( C->varType[v] == type ) {
	if( C->lower[v] > C->upper[v] ) {
	    return( FALSE );
	}
	if( C->lower[v] == C->upper[v] ) {
	    C->filled[C->lower[v]]++;
	    if( C->filled[C->lower[v]] > width ) {
		return( FALSE );
	    }
	}
    }

    for( v = 1; v <= C->n; v++ ) if( C->varType[v] == type ) {
	if( C->lower[v] < C->upper[v] ) {
	    if( C->filled[C->lower[v]] == width ) {
		while( C->filled[C->lower[v]] == width ) {
		    C->lower[v]++;
		}
		if( C->lower[v] > C->upper[v] ) {
		    return( FALSE );
		}
		push( v, C->s_lower );
		C->changes++;
		C->iterNeeded[C->varType[v]] = 1;
	    }
	    if( C->filled[C->upper[v]] == width ) {
		while( C->filled[C->upper[v]] == width ) {
		    C->upper[v]--;
		}
		if( C->lower[v] > C->upper[v] ) {
		    return( FALSE );
		}
		push( v, C->s_upper );
		C->changes++;
		C->iterNeeded[C->varType[v]] = 1;
	    }
	}
    }

    return( TRUE );
}

int mincmpiv(const void *x, const void *y)
{
    return (* ((interval **)x))->min - (* ((interval **)y))->min;
}

int maxcmpiv(const void *x, const void *y)
{
    return (* ((interval **)x))->max - (* ((interval **)y))->max;
}

static void
sortit( CSP_type *C )
{
    int i,j,nb,min,max,last,niv=C->niv;

    for (i=0; i<niv; i++) {
	C->minsorted[i] = C->maxsorted[i] = &C->iv[i];
    }

    qsort(C->minsorted, niv, sizeof(interval *), mincmpiv);
    qsort(C->maxsorted, niv, sizeof(interval *), maxcmpiv);

    min = C->minsorted[0]->min;
    max = C->maxsorted[0]->max + 1;
    C->bounds[0] = last = min-2;

    for (i=j=nb=0;;) { // merge minsorted[] and maxsorted[] into bounds[]
        if (i<niv && min<=max) {	// make sure minsorted exhausted first
            if (min != last)
                C->bounds[++nb] = last = min;
            C->minsorted[i]->minrank = nb;
            if (++i < niv)
                min = C->minsorted[i]->min;
        } else {
            if (max != last)
                C->bounds[++nb] = last = max;
            C->maxsorted[j]->maxrank = nb;
            if (++j == niv) break;
            max = C->maxsorted[j]->max + 1;
        }
    }
    C->nb = nb;
    C->bounds[nb+1] = C->bounds[nb] + 2;
}

/*
 *  set t[start],t[t[start]]... up to but excluding t[end], to to
 */
static void
pathset( int *t, int start, int end, int to )
{
    int k, l;

    for (l=start; (k=l) != end; t[k]=to) {
        l = t[k];
    }
}

static int
pathmin( int *t, int i )
{
    for (; t[i] < i; i=t[i])
	;
    return i;
}

static int
pathmax( int *t, int i )
{
    for (; t[i] > i; i=t[i])
	;
    return i;
}

static BOOL
filterlower( CSP_type *C, int *t, int *d, int *h, int width)
{
  int i,j,w,x,y,z;

    for (i=1; i<=C->nb+1; i++) {
        t[i] = h[i] = i-1;
        d[i] = width*(C->bounds[i] - C->bounds[i-1]);
    }
    for (i=0; i<C->niv; i++) { // visit intervals in increasing max order
        x = C->maxsorted[i]->minrank;
        y = C->maxsorted[i]->maxrank;
        j = t[z = pathmax(t, x+1)];
        if (--d[z] == 0) {
            t[z = pathmax(t, t[z]=z+1)] = j;
        }
        pathset(t, x+1, z, z); // path compression
        if (d[z] < width*(C->bounds[z]-C->bounds[y])) {
	    return( FALSE ); // no solution
        }
        if (h[x] > x) {
            C->maxsorted[i]->min = C->bounds[w = pathmax(h, h[x])];
            pathset(h, x, w, w); // path compression
        }
        if (d[z] == width*(C->bounds[z]-C->bounds[y])) {
            pathset(h, h[y], j-1, y); // mark hall interval
            h[y] = j-1; //("hall interval [%d,%d)\n",C->bounds[j],C->bounds[y]);
        }
    }
    return( TRUE );
}

static BOOL
filterupper( CSP_type *C, int *t, int *d, int *h, int width)
{
    int i,j,w,x,y,z;

    for (i=0; i<=C->nb; i++) {
        t[i] = h[i] = i+1;
        d[i] = width*(C->bounds[i+1] - C->bounds[i]);
    }
    for (i=C->niv; --i>=0; ) { // visit intervals in decreasing min order
        x = C->minsorted[i]->maxrank;
        y = C->minsorted[i]->minrank;
        j = t[z = pathmin(t, x-1)];
        if (--d[z] == 0) {
            t[z = pathmin(t, t[z]=z-1)] = j;
        }
        pathset(t, x-1, z, z);
        if (d[z] < width*(C->bounds[y]-C->bounds[z])) {
	    return( FALSE );
        }
        if (h[x] < x) {
            C->minsorted[i]->max = C->bounds[w = pathmin(h, h[x])] - 1;
            pathset(h, x, w, w);
        }
        if (d[z] == width*(C->bounds[y]-C->bounds[z])) {
            pathset(h, h[y], j+1, y);
            h[y] = j+1;
        }
    }
    return( TRUE );
}

/*
 *  Propagate the gcc constraint.
 */
static BOOL
propagateGCC( CSP_type *C, int type, int width )
{
    int  i, n;
    BOOL consistent;

    n = 0;
    for( i = 1; i <= C->n; i++ ) if( C->varType[i] == type ) {
	if( C->lower[i] > C->upper[i] ) {
	    return( FALSE );
	}
	C->iv[n].min = C->lower[i];
	C->iv[n].max = C->upper[i];
	n++;
    }
    if( n <= width ) {
	return( TRUE );
    }

    C->niv = n;

    sortit(C);

    consistent =
	filterlower( C, C->t, C->d, C->h, width ) &&
	filterupper( C, C->t, C->d, C->h, width );

    if( consistent ) {
	n = 0;
	for( i = 1; i <= C->n; i++ ) if( C->varType[i] == type ) {
	    if( C->lower[i] < C->iv[n].min ) {
		C->lower[i] = C->iv[n].min;
		push( i, C->s_lower );
		C->changes++;
	    }
	    if( C->upper[i] > C->iv[n].max ) {
		C->upper[i] = C->iv[n].max;
		push( i, C->s_upper );
		C->changes++;
	    }
	    n++;
	}
    }

    return( consistent );
}

static BOOL
safePruneUpper( CSP_type *C, int type, int width )
{
    int  i, j, k, a, b, c, d, count, l;
    BOOL changed, okToPrune;
    ELEM *p;

    count = 0;
    for( i = 1; i <= C->n; i++ ) if( C->varType[i] == type ) {
	C->values[count].bound = C->lower[i];
	C->values[count].data  = i;
	count++;
    }

    if( count == 0 ) {
	return( FALSE );
    }

    sortmin( C->values, count );

    i = 0;
    changed = FALSE;
    while( i < count ) {
	a = C->values[i].bound;
	/*
	 *  Step through all variables with lower bound equal to "a".
	 */
	okToPrune = TRUE;
	for( j = i; (j < count) && (a == C->values[j].bound); j++ ) {
	    if( okToPrune ) {
		/*
		 *  Check whether variable C->values[j].data satisfies
		 *  condition (ii) of safe pruning theorem.
		 */
		for( p = C->latencyPred[C->values[j].data];
			okToPrune && (p != NULL); p = p->next ) {
		    if( (C->upper[p->vertex]+p->latency) >
				C->lower[C->values[j].data] ) {
			okToPrune = FALSE;
		    }
		}
	    }
	}
	if( okToPrune ) {
	    l = j-i;
	    b = (int)ceil( (double)l/(double)width + (double)a - 1.0 );
	    /*
	     *  If there is a variable of type t with domain [c, d]
	     *  that intersects [a,b] but c != a, then not okToPrune.
	     */
	    for( k = 1; (k <= C->n) && okToPrune; k++ ) if( C->varType[k] == type ) {
		c = C->lower[k];
		d = C->upper[k];
		if( (c < a && d >= a) || (c > a && c <= b) ) {
		    okToPrune = FALSE;
		}
	    }
	}
	if( okToPrune ) {
	    /*
	     *  Prune variables C->values[i].data .. C->values[j-1].data.
	     */
	    for( k = i; k < j; k++ ) {
		if( C->upper[C->values[k].data] > b ) {
		    C->upper[C->values[k].data] = b;
		    if( a == b ) changed = TRUE;
		}
	    }
	}
	i = j;
    }

    return( changed );
}

static BOOL
safePruneLower( CSP_type *C, int type, int width )
{
    int  i, j, k, a, b, c, d, count, l;
    BOOL changed, okToPrune;
    ELEM *p;

    count = 0;
    for( i = 1; i <= C->n; i++ ) if( C->varType[i] == type ) {
	C->values[count].bound = C->upper[i];
	C->values[count].data  = i;
	count++;
    }

    if( count == 0 ) {
	return(FALSE);
    }

    sortmax( C->values, count );

    i = 0;
    changed = FALSE;
    while( i < count ) {
	b = C->values[i].bound;
	/*
	 *  Step through all variables with upper bound equal to "b".
	 */
	okToPrune = TRUE;
	for( j = i; (j < count) && (b == C->values[j].bound); j++ ) {
	    if( okToPrune ) {
		/*
		 *  Check whether variable C->values[j].data satisfies
		 *  condition (ii) of safe pruning theorem.
		 */
		for( p = C->latencySucc[C->values[j].data];
			okToPrune && (p != NULL); p = p->next ) {
		    if( (C->upper[C->values[j].data]) >
				C->lower[p->vertex]-p->latency ) {
			okToPrune = FALSE;
		    }
		}
	    }
	}
	if( okToPrune ) {
	    l = j-i;
	    a = (int)floor( (double)b + 1.0 - (double)l/(double)width );
	    /*
	     *  If there is a variable of type t with domain [c, d]
	     *  that intersects [a,b] but d != b, then not okToPrune.
	     */
	    for( k = 1; (k <= C->n) && okToPrune; k++ ) if( C->varType[k] == type ) {
		c = C->lower[k];
		d = C->upper[k];
		if( (d < b && d >= a) || (d > b && c <= b) ) {
		    okToPrune = FALSE;
		}
	    }
	}
	if( okToPrune ) {
	    /*
	     *  Prune variables C->values[i].data .. C->values[j-1].data.
	     */
	    for( k = i; k < j; k++ ) {
		if( C->lower[C->values[k].data] < a ) {
		    C->lower[C->values[k].data] = a;
		    if( a == b ) changed = TRUE;
		}
	    }
	}
	i = j;
    }

    return( changed );
}


/*
 *  Correctness relies on latency and gcc constraints
 *  having been propagated prior to calling. No need to put
 *  inside main propagator loop as changes here cannot
 *  cause further pruning by latency and gcc constraints.
 */
void
safePrune( CSP_type *C )
{
    int  type;

    for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
	if( C->nOps[type] > 0 ) {
	    while( safePruneUpper( C, type, nFunctionalUnits( type ) ) )
		;
	    while( safePruneLower( C, type, nFunctionalUnits( type ) ) )
		;
	}
    }
}

