
#define DEBUG 0

#include "global.h"
#include "processor.h"
#include "list.h"
#include "dag.h"
#include "csp.h"
#include "solve.h"

/*
 *  Choose the next variable to be branched on.
 */
static int
getNextVariable( CSP_type *C, int currentLevel )
{
    int    i, domainSize, minDomainSize, minVariable, maxChanges;

    maxChanges = 0;
    if( C->propagationLevel > 0 ) {
	for( i = 1; i <= C->n; i++ ) if( C->upper[i] > C->lower[i] ) {
	    if( maxChanges < C->changev[i] ) {
		maxChanges = C->changev[i];
	    }
	}
    }

    minDomainSize = HUGE;
    minVariable = 0;
    for( i = 1; i <= C->n; i++ ) if( C->upper[i] > C->lower[i] ) {

	domainSize = C->upper[i] - C->lower[i] + 1;

	/*
	 *  0.80 appears to be a good value.
	 *  If propagationLevel == 0, all variables will be considered.
	 *  If propagationLevel >  0, only variables near enough to max
	 *      changes will be considered (there will be at least one).
	 */
	if( C->changev[i] >= (0.8*maxChanges) ) {

	    if( (domainSize < minDomainSize) ) {
		minDomainSize = domainSize;
		minVariable = i;
	    }
	    else
	    if( (domainSize == minDomainSize) &&
		(C->nPreds[i] > C->nPreds[minVariable]) ) {
		minVariable = i;
	    }
	    else
	    if( (domainSize == minDomainSize) &&
		(C->nPreds[i] == C->nPreds[minVariable]) &&
		(C->nSuccs[i] > C->nSuccs[minVariable]) ) {
		minVariable = i;
	    }
	}
    }

    return( minVariable );
}

/*
 *  Perform singleton consistency to max depth.
 */
static int
singletonConsistency( CSP_type *C, int currentLevel, int depth, int max_depth )
{
    int  a, b, i, domainSize;
    BOOL consistent, changed;
int iter;

    saveBounds( C, currentLevel+depth );

    for( i = 1; i <= C->n; i++ ) {
	C->changev[i] = 0;
    }

    for( iter = 0; iter <= 1; iter++ )
    for( i = 1; i <= C->n; i++ ) if( (iter==0 && C->nPreds[i]>1)
					|| (iter==1 && C->nPreds[i]<=1) ){

	domainSize = C->upper[i] - C->lower[i] + 1;
	if( domainSize == 1 ) {
	    continue;
	}

	changed = FALSE;

#if DEBUG
	if( depth==0 )
	    printf("sc: variable %d: [%d, %d], currentLevel = %d\n",
			i,C->lower[i],C->upper[i],currentLevel);
#endif

	if( limit_exceeded( C ) ) {
	    return( TIMED_OUT );
	}

	a = C->lower[i];
	b = C->upper[i];

	consistent = FALSE;
	for( ; a <= b && !consistent; a++ ) {
	    C->lower[i] = a;
	    C->upper[i] = a;
	    consistent = propagate( C, i, currentLevel );
	    C->changev[i] = C->changes;
	    if( consistent && (depth < max_depth) ) {
		consistent = singletonConsistency( C, currentLevel, depth+1, max_depth );
	    }
	    restoreBounds( C, currentLevel+depth );
	    if( consistent && (C->lower[i] != a) ) {
	        C->lower[i] = a;
		changed = TRUE;
	        saveBounds( C, currentLevel+depth );
	    }
	}
	if( !consistent || (C->lower[i] > C->upper[i]) ) {
	    return( NO_SOLUTION );
	}

	consistent = FALSE;
	for( ; b >= a && !consistent; b-- ) {
	    C->lower[i] = b;
	    C->upper[i] = b;
	    consistent = propagate( C, i, currentLevel );
	    if( consistent && (depth < max_depth) ) {
		consistent = singletonConsistency( C, currentLevel, depth+1, max_depth );
	    }
	    restoreBounds( C, currentLevel+depth );
	    if( consistent && (C->upper[i] != b) ) {
	        C->upper[i] = b;
		changed = TRUE;
	        saveBounds( C, currentLevel+depth );
	    }
	}

	if( !consistent && (a > b) && (C->upper[i] > C->lower[i]) ) {
	    C->upper[i] = C->lower[i];
	    changed = TRUE;
	    saveBounds( C, currentLevel+depth );
	}
	if( C->lower[i] > C->upper[i] ) {
	    return( NO_SOLUTION );
	}
#if DEBUG
	if( depth==0 )
	    printf("    variable %d: [%d, %d], currentLevel = %d\n",
		i,C->lower[i],C->upper[i],currentLevel);
#endif
	if( changed ) {
	    if( !propagate( C, i, currentLevel ) ) {
		return( NO_SOLUTION );
	    }
	    saveBounds( C, currentLevel+depth );
	}
    }

    return( SOLUTION );
}

/*
 *  Interface to singletonConsistency routine.
 *  Returns SOLUTION if domains are non-empty after probing.
 *  Returns NO_SOLUTION if domains are empty after probing.
 *  Returns TIMED_OUT if time expired before probing was complete.
 */
static int
probe( CSP_type *C, int currentLevel )
{
    return( singletonConsistency( C, currentLevel,
		/*depth=*/0, /*max_depth*/C->propagationLevel-1 ) );
}


/*
 *  Solve the CSP by backtracking search interleaved with
 *  constraint propagation and chronological backtracking.
 *  Returns SOLUTION if a solution has been found for the CSP.
 *  Returns NO_SOLUTION if CSP has been proven inconsistent.
 *  Returns TIMED_OUT if time expired before backtracking was complete.
 */
int
backtrack( CSP_type *C, int currentLevel )
{
    int  a, b, currentVar, status;

    saveBounds( C, currentLevel );

    if( C->propagationLevel > 0 ) {
	safePrune( C );
	status = probe( C, currentLevel );
	if( status != SOLUTION ) {
	    return( status );
	}
    }

    currentVar = getNextVariable( C, currentLevel );
    if( currentVar == 0 ) {
        return( SOLUTION );
    }
#if DEBUG
    if( C->propagationLevel > 0 )
	printf("%d: next variable at level %d is %d, [%d, %d], %d\n",
		C->n,currentLevel,currentVar,
		C->lower[currentVar],
		C->upper[currentVar],
		C->varType[currentVar]);
#endif

    a = C->lower[currentVar];
    b = C->upper[currentVar];

    for( ; a <= b; a++ ) {

	C->lower[currentVar] = a;
	C->upper[currentVar] = a;
#if DEBUG
    if( C->propagationLevel > 0 )
	printf("\t%d <-- %d, remaining: [%d,%d], level: %d\n",currentVar,a,a+1,b,currentLevel);
#endif

	C->visits++;
//TEMPORARY && (C->backtracks > 10)
	if( limit_exceeded( C ) && (C->backtracks > 10) ) {
	    return( TIMED_OUT );
	}

        if( propagate( C, currentVar, currentLevel ) ) {
            if( (status = backtrack( C, currentLevel+1 )) ) {
                return( status );
            }
        }

        restoreBounds( C, currentLevel );
    }

    C->backtracks++;

    return( NO_SOLUTION );
}

/*
 *  Interface to backtracking routine.
 *  Returns SOLUTION,  length is length of optimal solution.
 *  Returns TIMED_OUT, length is lower bound on length of optimal solution.
 */
int
solve( CSP_type *C, int *length )
{
    int	 status;

    saveBounds( C, 0 );

    while( 1 ) {
	*length = C->m;

	/*
	 *  If the number of cycles needed to schedule the instructions
	 *  is equal to the known upper bound on the schedule, we're done.
	 */
	if( C->m == C->ub ) {
	    return( PROVED_OPTIMAL );
	}

#if DEBUG
	printf("%d: attempting schedule length %d\n",C->n,C->m);
#endif
	if( propagate( C, 1, /*currentLevel=*/0 ) ) status = backtrack( C, 1 );
	else					    status = NO_SOLUTION;

	if( status == SOLUTION ) {
	    return( SOLUTION );
	}

	if( status == NO_SOLUTION ) {
	    restoreBounds( C, 0 );
	    incrementUpperBounds( C );
	    saveBounds( C, 0 );
	}

	if( status == TIMED_OUT ) {

	    if( !C->restartsAllowed ) {
		return( TIMED_OUT );
	    }
	    if( (C->n <= MAX_RESTART_SIZE) && (C->propagationLevel == 0) ) {
		set_relative_CPU_limit( C, TIME_LIMIT_SECOND );
		C->propagationLevel++;
		restoreBounds( C, 0 );
	        status = NO_SOLUTION;
	    }
	    else
	    if( (C->n <= MAX_RESTART_SIZE) && (C->propagationLevel == 1) ) {
		set_absolute_CPU_limit( C, TIME_LIMIT_TOTAL );
		C->propagationLevel++;
		restoreBounds( C, 0 );
	        status = NO_SOLUTION;
	    }
	    else
	    if( (C->n >  MAX_RESTART_SIZE) && (C->propagationLevel == 0) ) {
		set_absolute_CPU_limit( C, TIME_LIMIT_TOTAL );
		C->propagationLevel++;
		restoreBounds( C, 0 );
	        status = NO_SOLUTION;
	    }
	    else {
		return( TIMED_OUT );
	    }

	}

    };
}

int
optimalScheduler( DAG_type *DAG, int order[] )
{
    CSP_type    *C;
    int		i, type, length, status;

    C = createCSP( DAG->nInstructions );
    for( i = 1; i <= C->n; i++ ) {
	C->varType[i] = DAG->instructionType[i-1];
    }
    for( type = 0; type < N_INSTRUCTION_TYPES; type++ ) {
	C->nOps[type] = 0;
    }
    for( i = 1; i <= C->n; i++ ) {
	C->nOps[C->varType[i]]++;
	C->changev[i] = 0;
    }

    setConstraints( C, DAG );
    setBounds( C );

    /*
     *  Determine an upper bound on the optimal schedule.
     */
    C->ub = listScheduler( DAG, order );

    set_relative_CPU_limit( C, TIME_LIMIT_FIRST );
    C->propagationLevel = 0;
    C->restartsAllowed = TRUE;
    status = solve( C, &length );

    if( status == SOLUTION ) {
        verifySolution( C );
        // Overwrite the solution found by the list scheduler
        for( i = 1; i <= C->n; i++ ) {
	    order[i-1] = C->lower[i];
        }
    }

    printf("%s: n: %d, LS: %d, CP: %d, v: %d, b: %d, s: %d, t: %0.2f\n",
	DAG->blockID, DAG->nInstructions, C->ub, length, C->visits,
	C->backtracks, status, elapsed_time( C ) );
    fflush( stdout );

    destroyCSP( C );

    return( length );
}

